/*
 * Created on May 29, 2005
 */
package com.simutronics.eclipsefe;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.URL;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.bindings.keys.KeyStroke;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.PreferenceConverter;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ST;
import org.eclipse.swt.custom.StyleRange;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.custom.VerifyKeyListener;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.MenuAdapter;
import org.eclipse.swt.events.MenuEvent;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.VerifyEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.ColorDialog;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.FontDialog;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.ScrollBar;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IViewSite;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.keys.IBindingService;
import org.eclipse.ui.part.ViewPart;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.simutronics.eclipsefe.extensions.ICommandProcessor;
import com.simutronics.eclipsefe.extensions.IInputProcessor;
import com.simutronics.eclipsefe.extensions.IMacroProcessor;
import com.simutronics.eclipsefe.extensions.IServiceProvider;
import com.simutronics.eclipsefe.preferences.HighlightsPreferencePage;
import com.simutronics.eclipsefe.preferences.MacroPreferencePage;
import com.simutronics.eclipsefe.preferences.StoredInfoPreferencePage;
import com.simutronics.eclipsefe.preferences.VariablesPreferencePage;

import static com.simutronics.eclipsefe.preferences.FEPreferencePage.*;

/**
 * @author Adam Vartanian
 */
public class FEView extends ViewPart {

	private enum Status { LOGIN, GAME_SELECT, CHARACTER_SELECT, CONNECTING, CONNECTED, DISCONNECTED }
	
	public static final String VIEWID = "com.simutronics.eclipsefe.feview";
	private static int viewnumber = 0;

	private Composite parentComposite;
	private Composite loginScreen;
	private Composite gameScreen;
	private Text account;
	private Text password;
	private Label gameSelectLabel;
	private Combo gameSelection;
	private Label charSelectLabel;
	private Combo charSelection;
	private Button back;
	private Button forward;
	private Label errorText;
	private SocketChannel sc;
	private String charID = null;
	
	int timeOffset = 0;
	
	Text commandEntry;
	Status status;
	private Canvas injuries;
	
	private Map<String, Label> progressBars = new HashMap<String, Label>();
	private Map<String, Label> labels = new HashMap<String, Label>();
	private Map<String, StyledText> windows = new HashMap<String, StyledText>();
	private Map<String, String> closedWindowRouting = new HashMap<String, String>();
	private ArrayList<Composite> composites = new ArrayList<Composite>();
	private StyledText gameView;
	private Socket accessConnection;
	private BufferedReader accessIn;
	private PrintStream accessOut;
	private Socket gameConnection;
	//private PrintStream output;
	private Color metabackground;
	private Map<String, Color> presets = new HashMap<String, Color>();
	private boolean localecho;
	
	private int bold = 0;
	private int boldstart = 0;
	private int presetstart = 0;
	private String stream = "main";
	private String streamThisLine = "main";
	boolean playmode = true;
	private boolean inComponent = false;
	private boolean inLeft = false;
	private boolean inRight = false;
	private boolean inSpell = false;
	private boolean inInv = false;
	private boolean prompt = false;
	private boolean addedPrompt = false;
	private String currentPreset = null;
	private String currentCompdef = null;
	private StringBuilder componentBuilder = new StringBuilder();
	private Map<String, Character> entities = new HashMap<String, Character>();
	
	private IPreferenceStore prefs;
	
	private Map<String, ArrayList<StyleRange>> presetRanges = new HashMap<String, ArrayList<StyleRange>>();
	
	private IBindingService bindingService;
	private Map<KeyStroke, String> macros;
	private Map<String, Highlight> highlights;
	private VariableService variables;
	
	private Map<String, Boolean> trailingNewlines = new HashMap<String, Boolean>();
	List<String> sentCommands = new ArrayList<String>();
	int currentCommand = -1;
	
	public FEView() {
		super();
		entities.put("amp", '&');
		entities.put("quot", '"');
		entities.put("lt", '<');
		entities.put("gt", '>');
		entities.put("tilde", '~');
		entities.put("nbsp", ' ');
		entities.put("apos", '\'');
	}
	
	public void init(IViewSite site) throws PartInitException {
		super.init(site);
		bindingService = (IBindingService) site.getWorkbenchWindow().getWorkbench().getAdapter(IBindingService.class);
		prefs = EclipsefePlugin.getDefault().getPreferenceStore();
		prefs.addPropertyChangeListener(new IPropertyChangeListener(){
			public void propertyChange(PropertyChangeEvent e) {
				if (status.compareTo(Status.CONNECTING) >= 0) {
					if ((gameView != null) && !gameView.isDisposed()) {
						if (e.getProperty().equals(PREF_BACKGROUND_COLOR) || e.getProperty().equals(PREF_FOREGROUND_COLOR) || e.getProperty().equals(PREF_META_BACKGROUND_COLOR)) {
							refreshMetaColors();
						} else if (e.getProperty().equals(PREF_MAIN_FONT)) {
							refreshFonts();
						} else if (e.getProperty().equals(MacroPreferencePage.PREF_MACROS) && (charID != null)) {
							loadMacros(charID);
						} else if (e.getProperty().startsWith(WINDOW_PREFIX)) {
							refreshColor(e.getProperty());
						} else if (e.getProperty().startsWith(PRESET_PREFIX)) {
							loadPreset(e.getProperty().substring(PRESET_PREFIX.length()));
						} else if (e.getProperty().equals(HighlightsPreferencePage.HIGHLIGHT_MODIFY)) {
							Highlight current = (Highlight) e.getOldValue();
							Highlight modified = (Highlight) e.getNewValue();
							//modifyHighlight(current, modified);
							removeHighlight(current);
							addHighlight(modified);
						} else if (e.getProperty().equals(HighlightsPreferencePage.HIGHLIGHT_ADD)) {
							Highlight h = (Highlight) e.getNewValue();
							addHighlight(h);
						} else if (e.getProperty().equals(HighlightsPreferencePage.HIGHLIGHT_REMOVE)) {
							Highlight h = (Highlight) e.getOldValue();
							removeHighlight(h);
						} else if (e.getProperty().equals(PREF_LOCALECHO)) {
							localecho = (Boolean) e.getNewValue();
						} else if (e.getProperty().equals(VariablesPreferencePage.PREF_VARIABLES)) {
							loadVariables(charID);
						}
					} else if (gameView != null) {
						EclipsefePlugin.getDefault().getPreferenceStore().removePropertyChangeListener(this);
					}
				}
			}
		});
	}

	public void createPartControl(Composite parent) {
		parentComposite = parent;
		parent.setLayout(new GridLayout(1, false));
		loginScreen = new Composite(parent, SWT.NONE);
		loginScreen.setLayoutData(new GridData());
		loginScreen.setLayout(new GridLayout(4, false));
		
		new Label(loginScreen, SWT.NONE).setText("Account:");
		account = new Text(loginScreen, SWT.BORDER);
		GridData tempGD = new GridData();
		tempGD.horizontalAlignment = SWT.FILL;
		tempGD.widthHint = 200;
		account.setLayoutData(tempGD);
		account.setText(prefs.getString(StoredInfoPreferencePage.PREF_SAVED_ACCOUNT));
		account.addSelectionListener(new SelectionAdapter(){
			public void widgetDefaultSelected(SelectionEvent e) {
				goForward();
			}
		});
		
		new Label(loginScreen, SWT.NONE).setText("Password:");
		password = new Text(loginScreen, SWT.BORDER | SWT.PASSWORD);
		tempGD = new GridData();
		tempGD.horizontalAlignment = SWT.FILL;
		tempGD.widthHint = 200;
		password.setLayoutData(tempGD);
		password.addSelectionListener(new SelectionAdapter(){
			public void widgetDefaultSelected(SelectionEvent e) {
				goForward();
			}
		});
		if (account.getText().length() != 0) {
			password.setFocus();
		}
		
		gameSelectLabel = new Label(loginScreen, SWT.NONE);
		gameSelectLabel.setText("Select Game:");
		gameSelectLabel.setVisible(false);
		
		gameSelection = new Combo(loginScreen, SWT.DROP_DOWN | SWT.READ_ONLY);
		GridData gameSelGD = new GridData();
		gameSelGD.grabExcessHorizontalSpace = true;
		gameSelGD.horizontalAlignment = SWT.FILL;
		gameSelection.setLayoutData(gameSelGD);
		gameSelection.setVisible(false);
		gameSelection.addSelectionListener(new SelectionAdapter(){
			public void widgetDefaultSelected(SelectionEvent e) {
				goForward();
			}
		});

		charSelectLabel = new Label(loginScreen, SWT.NONE);
		charSelectLabel.setText("Select Character:");
		charSelectLabel.setVisible(false);
		
		charSelection = new Combo(loginScreen, SWT.DROP_DOWN | SWT.READ_ONLY);
		GridData charSelGD = new GridData();
		charSelGD.grabExcessHorizontalSpace = true;
		charSelGD.horizontalAlignment = SWT.FILL;
		charSelection.setLayoutData(charSelGD);
		charSelection.setVisible(false);
		charSelection.addSelectionListener(new SelectionAdapter(){
			public void widgetDefaultSelected(SelectionEvent e) {
				goForward();
			}
		});
		
		back = new Button(loginScreen, SWT.PUSH);
		back.setText("Back");
		back.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent event) {
				goBackward();
			}
		});
		forward = new Button(loginScreen, SWT.PUSH);
		forward.setText("OK");
		forward.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent event) {
				goForward();
			}
		});
		
		new Label(loginScreen, SWT.NONE);
		new Label(loginScreen, SWT.NONE);
		
		errorText = new Label(loginScreen, SWT.NONE);
		errorText.setForeground(parent.getDisplay().getSystemColor(SWT.COLOR_RED));
		GridData errorGD = new GridData(SWT.FILL, SWT.NONE, true, false);
		errorGD.horizontalSpan = 4;
		errorText.setLayoutData(errorGD);
		
		status = Status.LOGIN;
		createActions();
	}
	
	private void createActions() {
		IMenuManager menu = getViewSite().getActionBars().getMenuManager();
		menu.add(new Action("New View"){
			public synchronized void run() {
				try {
					getViewSite().getWorkbenchWindow().getActivePage().showView(FEView.VIEWID, Integer.toString(viewnumber), IWorkbenchPage.VIEW_VISIBLE);
					viewnumber++;
				} catch (PartInitException e) {
					// do nothing, it didn't open
				}
			}
		});
	}
	
	private void error(String message) {
		errorText.setText(message);
	}
	
	private void goForward() {
		if (status == Status.LOGIN) {
			if (account.getText().trim().equals("")) {
				error("You must enter an account name.");
				return;
			}
			if (password.getText().trim().equals("")) {
				error("You must enter a password.");
				return;
			}
			try {
				error("Connecting to Simutronics...");
				errorText.redraw();
				errorText.update();
				accessConnection = new Socket("eaccess.play.net", 7900);
				accessIn = new BufferedReader(new InputStreamReader(accessConnection.getInputStream()));
				accessOut = new PrintStream(accessConnection.getOutputStream());
				accessOut.println("K");
				String passwordHash = accessIn.readLine();
				byte[] pHash = passwordHash.getBytes();
				if (pHash.length != 32) {
					errorText.setText("Invalid connection to Simutronics server (bad hash).");
					accessConnection.close();
					return;
				}
				byte[] pass = password.getText().getBytes();
				if (pass.length > 32) {
					error("Password too long.");
					accessConnection.close();
					return;
				}
				for (int i = 0; i < pass.length; i++) {
					pass[i] = (byte) (((pass[i] - 0x20) ^ pHash[i]) + 0x20);
				}
				String hashedPassword = new String(pass);
				accessOut.print("A\t" + account.getText() + "\t" + hashedPassword + "\r\n");
				String response = accessIn.readLine();
				if (!response.matches(".*\tKEY\t.*")) {
					error("Bad account name or password.");
					accessConnection.close();
					return;
				}
				accessOut.println("M");
				String instances = accessIn.readLine();
				ArrayList<String> instanceList = new ArrayList<String>();
				Collections.addAll(instanceList, instances.split("\t"));
				instanceList.remove(0);   // Get rid of the M response
				for (int i = 0; i < instanceList.size() / 2; i++) {
					accessOut.println("N\t" + instanceList.get(i*2));
					response = accessIn.readLine();
					response = response.split("\t")[1];
					if (response.matches(".*[| ]STORM.*")) {
						gameSelection.add(instanceList.get(i*2 + 1));
						gameSelection.setData(instanceList.get(i*2 + 1), instanceList.get(i*2));
					}
				}
				gameSelection.deselectAll();
				if (gameSelection.indexOf(prefs.getString(StoredInfoPreferencePage.PREF_SAVED_INSTANCE)) != -1) {
					gameSelection.select(gameSelection.indexOf(prefs.getString(StoredInfoPreferencePage.PREF_SAVED_INSTANCE)));	
				}
				gameSelection.setVisible(true);
				gameSelection.setEnabled(true);
				account.setEnabled(false);
				password.setEnabled(false);
				gameSelectLabel.setVisible(true);
				status = Status.GAME_SELECT;
				error("");
				if (!gameSelection.setFocus()) {
					forward.setFocus();
				}
			} catch (UnknownHostException e) {
				error("Could not connect to Simutronics server.");
			} catch (IOException e) {
				error("Could not connect to Simutronics server.");
			}
		} else if (status == Status.GAME_SELECT) {
			if (gameSelection.getSelectionIndex() == -1) {
				error("You must select a game.");
				return;
			}
			try {
				String selected = gameSelection.getItem(gameSelection.getSelectionIndex());
				selected = (String) gameSelection.getData(selected);
				accessOut.println("F\t" + selected);
				String response = accessIn.readLine().split("\t")[1];
				if (response.equals("EXPIRED") || response.equals("NEW_TO_GAME")) {
					error("No subscription to game.");
					return;
				}
				accessOut.println("G\t" + selected);
				accessIn.readLine();  // a bunch of website addresses and stuff
				accessOut.println("P\t" + selected);
				accessIn.readLine();  // the pricing information for this game
				accessOut.println("C");
				String[] chardata = accessIn.readLine().split("\t");  // the list of characters for this game
				ArrayList<String> charList = new ArrayList<String>();
				Collections.addAll(charList, chardata);
				for (int i = 0; i < 5; i++) {
					charList.remove(0);  // throw away C, numchars, numslots, something, something
				}
				charSelection.removeAll();
				for (int i = 0; i < charList.size() / 2; i++) {
					charSelection.add(charList.get(i*2 + 1));
					charSelection.setData(charList.get(i*2 + 1), charList.get(i*2));
				}
				charSelection.setVisible(true);
				charSelection.deselectAll();
				if (charSelection.indexOf(prefs.getString(StoredInfoPreferencePage.PREF_SAVED_CHARACTER)) != -1) {
					charSelection.select(charSelection.indexOf(prefs.getString(StoredInfoPreferencePage.PREF_SAVED_CHARACTER)));	
				}
				charSelectLabel.setVisible(true);
				gameSelection.setEnabled(false);
				error("");
				status = Status.CHARACTER_SELECT;
				if (!charSelection.setFocus()) {
					forward.setFocus();
				}
			} catch (IOException e) {
				error("Could not talk to Simutronics server.");
				status = Status.LOGIN;
				try {
					accessConnection.close();
				} catch (IOException e2) {
					// ignore
				}
				gameSelection.setVisible(false);
				gameSelectLabel.setVisible(false);
				account.setEnabled(true);
				password.setEnabled(true);
			}
		} else if (status == Status.CHARACTER_SELECT) {
			if (charSelection.getSelectionIndex() == -1) {
				error("You must select a character.");
				return;
			}
			try {
				String selection = (String) charSelection.getData(charSelection.getItem(charSelection.getSelectionIndex()));
				accessOut.println("L\t" + selection + "\tSTORM");
				String[] response = accessIn.readLine().split("\t");
				String host = null;
				String key = null;
				int port = 0;
				for (String line : response) {
					if (line.matches("GAMEHOST=.*")) {
						host = line.substring(9);
					} else if (line.matches("GAMEPORT=.*")) {
						port = Integer.parseInt(line.substring(9));
					} else if (line.matches("KEY=.*")) {
						key = line.substring(4);
					}
				}
				accessConnection.close();
				if (prefs.getBoolean(StoredInfoPreferencePage.PREF_AUTOSAVE_LOGIN_DATA)) {
					prefs.setValue(StoredInfoPreferencePage.PREF_SAVED_ACCOUNT, account.getText());
					prefs.setValue(StoredInfoPreferencePage.PREF_SAVED_CHARACTER, charSelection.getItem(charSelection.getSelectionIndex()));
					prefs.setValue(StoredInfoPreferencePage.PREF_SAVED_INSTANCE, gameSelection.getItem(gameSelection.getSelectionIndex()));
				}
				String gameName = gameSelection.getItem(gameSelection.getSelectionIndex());
				String gameID = (String) gameSelection.getData(gameName);
				charID = charSelection.getItem(charSelection.getSelectionIndex()) + "-" + gameID;
				setPartName(charSelection.getItem(charSelection.getSelectionIndex()));
				removeLoginScreen();
				setContentDescription(gameName);
				loadMacros(charID);
				loadHighlights(charID);
				loadVariables(charID);
				loadPresets();
				localecho = prefs.getBoolean(PREF_LOCALECHO);
				connect(host, port, key);
			} catch (IOException e) {
				error("Could not talk to Simutronics server.");
				status = Status.LOGIN;
				try {
					accessConnection.close();
				} catch (IOException e2) {
					// ignore
				}
				gameSelection.setVisible(false);
				gameSelectLabel.setVisible(false);
				charSelection.setVisible(false);
				charSelectLabel.setVisible(false);
				account.setEnabled(true);
				password.setEnabled(true);
			}
		}
	}
	
	private void goBackward() {
		if (status == Status.LOGIN) {
			return;
		} else if (status == Status.GAME_SELECT) {
			try {
				accessConnection.close();
			} catch (IOException e) {
				// whatever
			}
			gameSelection.setVisible(false);
			gameSelectLabel.setVisible(false);
			account.setEnabled(true);
			password.setEnabled(true);
			error("");
			status = Status.LOGIN;
		} else if (status == Status.CHARACTER_SELECT) {
			charSelection.setVisible(false);
			charSelectLabel.setVisible(false);
			gameSelection.setEnabled(true);
			status = Status.GAME_SELECT;
		}
	}
	
	private void removeLoginScreen() {
		if (loginScreen.isDisposed()) {
			if (EclipsefePlugin.DEBUG) {
				System.out.println("Widget is disposed, ignoring.");
			}
			return;
		}
		loginScreen.dispose();
		addGameScreen();
	}
	
	private void addGameScreen() {
		try {
			URL configurationLocation = EclipsefePlugin.pluginBundle.getEntry("/configuration/layout.xml");
			if (configurationLocation == null) {
				throw new FileNotFoundException("Could not find layout file.");
			}
			InputStream stream = configurationLocation.openStream();
			DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
			docBuilderFactory.setCoalescing(true);
			docBuilderFactory.setIgnoringComments(true);
			DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
			Document configuration = docBuilder.parse(stream);
			composites.add(parentComposite);
			buildLayout(configuration.getDocumentElement(), parentComposite);
		} catch (Exception e) {
			addWindow(gameScreen, "main");
			gameScreen.layout();
			displayError("Could not load layout configuration.");
			return;
		}
//		addWindow(gameScreen, "gmchat");
//		addWindow(gameScreen, "main");
		
//		Shell s = new Shell(parentComposite.getShell(), SWT.BORDER | SWT.CLOSE | SWT.MODELESS | SWT.RESIZE | SWT.TOOL);
//		s.setLayout(new GridLayout());
//		Label l = new Label(s, SWT.NONE);
//		l.setText("Bob bobbo bobobob");
//		s.layout();
//		s.open();
		
//		Composite statusBars = new Composite(gameScreen, SWT.NONE);
//		statusBars.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
//		statusBars.setLayout(new GridLayout(4, true));
//		statusBars.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
		
//		addProgressBar(statusBars, "health", Display.getCurrent().getSystemColor(SWT.COLOR_DARK_RED));
//		addProgressBar(statusBars, "mana", Display.getCurrent().getSystemColor(SWT.COLOR_BLUE));
//		addProgressBar(statusBars, "stamina", Display.getCurrent().getSystemColor(SWT.COLOR_DARK_YELLOW));
//		addProgressBar(statusBars, "spirit", Display.getCurrent().getSystemColor(SWT.COLOR_DARK_GRAY));

//		addCommandEntry(gameScreen);
		
		refreshMetaColors();
		commandEntry.setFocus();
		parentComposite.layout();
	}
	
	private void buildLayout(Element e, Composite parent) throws IOException {
		if (e.getTagName().equals("area")) {
			Composite c = addComposite(parent, e);
			NodeList children = e.getChildNodes();
			for (int i = 0; i < children.getLength(); i++) {
				Node n = children.item(i);
				if (n instanceof Element) {
					buildLayout((Element) n, c);
				}
			}
		} else if (e.getTagName().equals("progressBar")) {
			addProgressBar(parent, e);
		} else if (e.getTagName().equals("window")) {
			addWindow(parent, e);
		} else if (e.getTagName().equals("commandentry")) {
			addCommandEntry(parent, e);
		} else if (e.getTagName().equals("injuries")) {
			addInjuries(parent, e);
		} else if (e.getTagName().equals("label")) {
			addLabel(parent, e);
		}
	}
	
	private Composite addComposite(Composite parent, Element e) {
		Composite c = new Composite(parent, SWT.NONE);
		GridLayout gameLayout = new GridLayout();
		try {
			gameLayout.numColumns = Integer.parseInt(e.getAttribute("cols"));
		} catch (NumberFormatException ex) {
			gameLayout.numColumns = 1;
		}
		try {
			gameLayout.marginHeight = Integer.parseInt(e.getAttribute("marginheight"));
		} catch (NumberFormatException ex) {
			// ignore
		}
		try {
			gameLayout.marginWidth = Integer.parseInt(e.getAttribute("marginwidth"));
		} catch (NumberFormatException ex) {
			// ignore
		}
		gameLayout.makeColumnsEqualWidth = Boolean.parseBoolean(e.getAttribute("forceequalwidth"));
		c.setLayout(gameLayout);
		c.setLayoutData(getGridData(e));
		if (gameScreen == null) {
			gameScreen = c;
		}
		composites.add(c);
		return c;
	}
	
	private void addLabel(Composite parent, Element e) {
		Label l = new Label(parent, SWT.CENTER);
		l.setForeground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		l.setBackground(metabackground);
		l.setLayoutData(getGridData(e));
		l.setText("Waiting for data...");
		labels.put(e.getAttribute("id"), l);
	}
	
	/**
	 * Uses attributes:
	 * color - color of the bar
	 * id - id of the bar
	 * @param parent
	 * @param e
	 */
	private void addProgressBar(Composite parent, Element e) {
		Label l = new Label(parent, SWT.CENTER | SWT.BORDER);
		l.setForeground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		l.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
		l.setLayoutData(getGridData(e));
		l.setData("color", getColor(e.getAttribute("color")));
		String id = e.getAttribute("id");
		if (id.equals("roundtime") || id.equals("castrt")) {
			RoundtimeUpdater updater = new RoundtimeUpdater(l);
			l.setData("updater", updater);
			l.setData("roundtime", 0);
			l.addPaintListener(new RoundtimePaintListener());
			updater.start();
		} else {
			l.setData("percent", 0.0);
			l.setText("waiting for data...");
			l.addPaintListener(new StatusLabelPainter());
		}
		progressBars.put(id, l);
	}
	
	private void addInjuries(Composite parent, Element e) throws IOException {
		Composite c = new Composite(parent, SWT.NONE);
		c.setLayoutData(getGridData(e));
		c.setLayout(new GridLayout(2, false));
		
		injuries = new Canvas(c, SWT.NO_BACKGROUND | SWT.NO_FOCUS | SWT.NO_REDRAW_RESIZE);
		URL i = EclipsefePlugin.pluginBundle.getEntry("/images/injuriesman.png");
		if (i == null) {
			throw new FileNotFoundException("Could not find injuries graphic.");
		}
		InputStream imagestream = i.openStream();
		Image image = new Image(Display.getCurrent(), imagestream);
		injuries.setData("image", image);
		InjuriesPaintListener updater = new InjuriesPaintListener(injuries); 
		injuries.addPaintListener(updater);
		injuries.setData("updater", updater);
		GridData imageGD = new GridData(image.getBounds().width, image.getBounds().height);
		injuries.setLayoutData(imageGD);
		
		Group g = new Group(c, SWT.SHADOW_NONE);
		g.setText("Display");
		g.setLayout(new GridLayout(1, false));
		
		Button b = new Button(g, SWT.RADIO);
		b.setText("Injuries");
		b.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent ev) {
				if (((Button) ev.widget).getSelection()) {
					sendHiddenCommand("_injury 0");
				}
			}
		});
		injuries.setData("injrRad", b);
		b = new Button(g, SWT.RADIO);
		b.setText("Scars");
		b.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent ev) {
				if (((Button) ev.widget).getSelection()) {
					sendHiddenCommand("_injury 1");
				}
			}
		});
		injuries.setData("scarRad", b);
		b = new Button(g, SWT.RADIO);
		b.setText("Both");
		b.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent ev) {
				if (((Button) ev.widget).getSelection()) {
					sendHiddenCommand("_injury 2");
				}
			}
		});
		injuries.setData("bothRad", b);
	}
	
	private void addWindow(Composite parent, String id) {
		addWindow(parent, null, id);
	}
	
	private void addWindow(Composite parent, Element e) {
		addWindow(parent, e, e.getAttribute("id"));
	}
	
	private void addWindow(Composite parent, Element el, String id) {
		if (id.equals("commandentry")) {       // ID commandentry is not valid
			return;
		}
		StyledText st = new StyledText(parent, SWT.V_SCROLL | SWT.WRAP | SWT.MULTI | SWT.READ_ONLY | SWT.BORDER);
		if (el != null) {
			st.setLayoutData(getGridData(el));
		} else {
			st.setLayoutData(new GridData(GridData.FILL_BOTH));
		}
		st.setData("id", id);
		st.getVerticalBar().setVisible(false);
		st.addVerifyKeyListener(new VerifyKeyListener() {
			public void verifyKey(VerifyEvent e) {
				if ((0x20 <= e.character) && (e.character <= 0x7E) && ((e.stateMask == SWT.NONE) || (e.stateMask == SWT.SHIFT))) {
					commandEntry.append(Character.toString(e.character));
					commandEntry.setFocus();
					e.doit = false;
				}
			}
		});
//		st.addSelectionListener(new SelectionAdapter() {
//			public void widgetSelected(SelectionEvent e) {
//				commandEntry.setFocus();
//			}
//		});
		st.setMenu(getContextMenu(st));
		st.addControlListener(new ControlListener() {
			public void controlMoved(ControlEvent e) {
				((StyledText) e.widget).setSelection(((StyledText)e.widget).getCharCount());
			}
			
			public void controlResized(ControlEvent e) {
				((StyledText) e.widget).setSelection(((StyledText) e.widget).getCharCount());
			}
		});

		windows.put(id, st);
		
		PreferenceConverter.setDefault(prefs, WINDOW_PREFIX + id + FOREGROUND_SUFFIX, PreferenceConverter.getColor(prefs, PREF_FOREGROUND_COLOR));
		PreferenceConverter.setDefault(prefs, WINDOW_PREFIX + id + BACKGROUND_SUFFIX, PreferenceConverter.getColor(prefs, PREF_BACKGROUND_COLOR));
		PreferenceConverter.setDefault(prefs, WINDOW_PREFIX + id + FONT_SUFFIX, PreferenceConverter.getFontData(prefs, PREF_MAIN_FONT));
		
		st.setForeground(new Color(Display.getCurrent(), PreferenceConverter.getColor(prefs, WINDOW_PREFIX + id + FOREGROUND_SUFFIX)));
		st.setBackground(new Color(Display.getCurrent(), PreferenceConverter.getColor(prefs, WINDOW_PREFIX + id + BACKGROUND_SUFFIX)));
		st.setFont(new Font(Display.getCurrent(), PreferenceConverter.getFontData(prefs, WINDOW_PREFIX + id + FONT_SUFFIX)));
		st.addDisposeListener(new FontDisposer(st.getFont()));
		
		if (id.equals("main")) {
			gameView = st;
		}
	}
	
	private void addCommandEntry(Composite parent, Element e) {
		if (commandEntry != null) {
			return;
		}
		commandEntry = new Text(gameScreen, SWT.MULTI | SWT.LEFT | SWT.BORDER);
		commandEntry.setLayoutData(getGridData(e));
		commandEntry.addModifyListener(new ModifyListener(){
			public void modifyText(ModifyEvent e) {
				if (EclipsefePlugin.DEBUG) {
					System.out.println("ModifyEvent: text is now " + ((Text) e.widget).getText());
				}
				sendCommands();
			}
		});
		commandEntry.addKeyListener(new KeyAdapter(){
			public void keyPressed(KeyEvent e) {
				if (EclipsefePlugin.DEBUG) {
					System.out.print("A key was pressed: " + e.keyCode + " with statemask " + e.stateMask + ".  ");
					System.out.println("This keystroke would look like: " + KeyStroke.getInstance(e.stateMask, e.keyCode) + " or " + KeyStroke.getInstance(e.stateMask, e.keyCode).format());
				}
				if ((e.keyCode == SWT.CR) && (e.stateMask == 0)) {
					commandEntry.append("\n");
					e.doit = false;
				} else if (MacroPreferencePage.validMacroKey(e)) {
					e.doit = false;
					String result = macros.get(KeyStroke.getInstance(e.stateMask, e.keyCode));
					if (result == null) {
						// do nothing
					} else if (result.equalsIgnoreCase("{pageup}")) {
						gameView.invokeAction(ST.PAGE_UP);
					} else if (result.equalsIgnoreCase("{pagedown}")) {
						gameView.invokeAction(ST.PAGE_DOWN);
					} else if (result.equalsIgnoreCase("{lineup}")) {
						gameView.invokeAction(ST.WINDOW_START);
						gameView.invokeAction(ST.LINE_UP);
					} else if (result.equalsIgnoreCase("{linedown}")) {
						gameView.invokeAction(ST.WINDOW_END);
						gameView.invokeAction(ST.LINE_DOWN);
					} else if (result.equalsIgnoreCase("{cut}")) {
						commandEntry.cut();
					} else if (result.equalsIgnoreCase("{copy}")) {
						if (commandEntry.getSelectionCount() > 0) {
							commandEntry.copy();
						} else {
							gameView.copy();
						}
					} else if (result.equalsIgnoreCase("{paste}")) {
						commandEntry.paste();
					} else if (result.equalsIgnoreCase("{selectall}")) {
						commandEntry.selectAll();
					} else if (result.equalsIgnoreCase("{test}")) {
						// do nothing
					} else if (result.equalsIgnoreCase("{lastcommand}")) {
						if (currentCommand < (sentCommands.size() - 1)) {
							currentCommand++;
							commandEntry.setText(sentCommands.get(currentCommand));
							commandEntry.setSelection(commandEntry.getCharCount());
						}
					} else if (result.equalsIgnoreCase("{nextcommand}")) {
						currentCommand--;
						if (currentCommand < 0) {
							commandEntry.setText("");
							currentCommand = -1;
						} else {
							commandEntry.setText(sentCommands.get(currentCommand));
							commandEntry.setSelection(commandEntry.getCharCount());
						}
					} else if (result.equalsIgnoreCase("{sendlastcommand}")) {
						if (sentCommands.size() > 0) {
							sendCommand(sentCommands.get(0), false);
						}
					} else if (result.equalsIgnoreCase("{sendsecondtolastcommand}")) {
						if (sentCommands.size() > 1) {
							sendCommand(sentCommands.get(1), false);
						}
					} else if (EclipsefePlugin.getMacroProcessors().containsKey(result.toLowerCase())) {
						IMacroProcessor processor = EclipsefePlugin.getMacroProcessors().get(result.toLowerCase());
						processor.processMacro(result, FEView.this);
					} else {
						if (result.indexOf("\\?") > -1) {
							InputDialog dialog = new InputDialog(gameView.getShell(), "Script Prompt", result, "", null);
							dialog.open();
							if (dialog.getReturnCode() == Window.CANCEL) {
								return;
							}
							String replacement = dialog.getValue();
							result = result.replace("\\?", replacement);
						}
						result = result.replace("\\r", "\n");
						if (result.indexOf("\\x") > -1) {
							commandEntry.setText("");
						}
						commandEntry.insert(result);
						sendCommands();
					}
				}
			}
		});
		commandEntry.addFocusListener(new FocusListener(){
			public void focusLost(FocusEvent e) {
				bindingService.setKeyFilterEnabled(true);
			}
			public void focusGained(FocusEvent e) {
				bindingService.setKeyFilterEnabled(false);
			}
		});
		
		commandEntry.setData("id", "commandentry");
		
		PreferenceConverter.setDefault(prefs, WINDOW_PREFIX + "commandentry" + FOREGROUND_SUFFIX, PreferenceConverter.getColor(prefs, PREF_FOREGROUND_COLOR));
		PreferenceConverter.setDefault(prefs, WINDOW_PREFIX + "commandentry" + BACKGROUND_SUFFIX, PreferenceConverter.getColor(prefs, PREF_BACKGROUND_COLOR));
		PreferenceConverter.setDefault(prefs, WINDOW_PREFIX + "commandentry" + FONT_SUFFIX, PreferenceConverter.getFontData(prefs, PREF_MAIN_FONT));
		
		commandEntry.setForeground(new Color(Display.getCurrent(), PreferenceConverter.getColor(prefs, WINDOW_PREFIX + "commandentry" + FOREGROUND_SUFFIX)));
		commandEntry.setBackground(new Color(Display.getCurrent(), PreferenceConverter.getColor(prefs, WINDOW_PREFIX + "commandentry" + BACKGROUND_SUFFIX)));
		commandEntry.setFont(new Font(Display.getCurrent(), PreferenceConverter.getFontData(prefs, WINDOW_PREFIX + "commandentry" + FONT_SUFFIX)));
		commandEntry.addDisposeListener(new FontDisposer(commandEntry.getFont()));
		
		commandEntry.setMenu(getContextMenu(commandEntry));
	}
	
	private Menu getContextMenu(Control control) {
		Menu m = new Menu(control);
		m.setData("control", control);
		
		if (control instanceof StyledText) {
			MenuItem mi = new MenuItem(m, SWT.PUSH);
			mi.setText("Copy");
			mi.setData("control", control);
			mi.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent se) {
					StyledText st = (StyledText) se.widget.getData("control");
					st.invokeAction(ST.COPY);
				}
			});
			m.addMenuListener(new MenuAdapter() {
				public void menuShown(MenuEvent e) {
					StyledText st = (StyledText) e.widget.getData("control");
					MenuItem mi = (MenuItem) e.widget.getData("copymenuitem");
					if (st.getSelectionCount() > 0) {
						mi.setEnabled(true);
					} else {
						mi.setEnabled(false);
					}
				}
			});
			m.setData("copymenuitem", mi);
		}
		
		MenuItem mi = new MenuItem(m, SWT.PUSH);
		mi.setText("Font...");
		mi.setData("control", control);
		mi.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent se) {
				Control c = (Control) se.widget.getData("control");
				FontDialog fd = new FontDialog(c.getShell(), SWT.APPLICATION_MODAL);
				fd.setText("Select window font");
				fd.setFontList(c.getFont().getFontData());
				FontData data = fd.open();
				if (data == null) {
					return;
				}
				Font oldFont = c.getFont();
				c.setFont(new Font(c.getDisplay(), data));
				oldFont.dispose();
				PreferenceConverter.setValue(EclipsefePlugin.getDefault().getPreferenceStore(), WINDOW_PREFIX + c.getData("id") + FONT_SUFFIX, data);
			}
		});
		
		mi = new MenuItem(m, SWT.PUSH);
		mi.setText("Foreground color...");
		mi.setData("control", control);
		mi.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent se) {
				Control c = (Control) se.widget.getData("control");
				ColorDialog cd = new ColorDialog(c.getShell(), SWT.APPLICATION_MODAL);
				cd.setText("Select foreground color");
				cd.setRGB(c.getForeground().getRGB());
				RGB data = cd.open();
				if (data == null) {
					return;
				}
				c.getForeground().dispose();
				c.setForeground(new Color(c.getDisplay(), data));
				PreferenceConverter.setValue(EclipsefePlugin.getDefault().getPreferenceStore(), WINDOW_PREFIX + c.getData("id") + FOREGROUND_SUFFIX, data);
			}
		});
		
		mi = new MenuItem(m, SWT.PUSH);
		mi.setText("Background color...");
		mi.setData("control", control);
		mi.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent se) {
				Control c = (Control) se.widget.getData("control");
				ColorDialog cd = new ColorDialog(c.getShell(), SWT.APPLICATION_MODAL);
				cd.setText("Select background color");
				cd.setRGB(c.getBackground().getRGB());
				RGB data = cd.open();
				if (data == null) {
					return;
				}
				c.getBackground().dispose();
				c.setBackground(new Color(c.getDisplay(), data));
				PreferenceConverter.setValue(EclipsefePlugin.getDefault().getPreferenceStore(), WINDOW_PREFIX + c.getData("id") + BACKGROUND_SUFFIX, data);
			}
		});
		
		if (control instanceof StyledText) {
			mi = new MenuItem(m, SWT.CHECK);
			mi.setText("Timestamps");
			mi.setData("control", control);
			mi.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent se) {
					Control c = (Control) se.widget.getData("control");
					boolean selected = ((MenuItem) se.widget).getSelection();
					c.setData("timestamps", selected);
					EclipsefePlugin.getDefault().getPreferenceStore().setValue(WINDOW_PREFIX + c.getData("id") + TIMESTAMP_SUFFIX, selected);
				}
			});
			boolean selected = EclipsefePlugin.getDefault().getPreferenceStore().getBoolean(WINDOW_PREFIX + control.getData("id") + TIMESTAMP_SUFFIX);
			mi.setSelection(selected);
			control.setData("timestamps", selected);
		}
		
		return m;
	}
	
	private Color getColor(String colorName) {
		if (colorName.equals("black")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_BLACK);
		} else if (colorName.equals("blue")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_BLUE);
		} else if (colorName.equals("cyan")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_CYAN);
		} else if (colorName.equals("dark blue")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_DARK_BLUE);
		} else if (colorName.equals("dark cyan")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_DARK_CYAN);
		} else if (colorName.equals("dark gray")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_DARK_GRAY);
		} else if (colorName.equals("dark green")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_DARK_GREEN);
		} else if (colorName.equals("dark magenta")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_DARK_MAGENTA);
		} else if (colorName.equals("dark red")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_DARK_RED);
		} else if (colorName.equals("dark yellow")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_DARK_YELLOW);
		} else if (colorName.equals("gray")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_GRAY);
		} else if (colorName.equals("green")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_GREEN);
		} else if (colorName.equals("magenta")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_MAGENTA);
		} else if (colorName.equals("red")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_RED);
		} else if (colorName.equals("white")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_WHITE);
		} else if (colorName.equals("yellow")) {
			return Display.getCurrent().getSystemColor(SWT.COLOR_YELLOW);
		}
		return Display.getCurrent().getSystemColor(SWT.COLOR_BLACK);
	}
	
	private GridData getGridData(Element e) {
		GridData gd = new GridData();
		String fill = e.getAttribute("fill");
		if (fill.equals("both") || fill.equals("horizontal")) {
			gd.horizontalAlignment = SWT.FILL;
			gd.grabExcessHorizontalSpace = true;
		}
		if (fill.equals("both") || fill.equals("vertical")) {
			gd.verticalAlignment = SWT.FILL;
			gd.grabExcessVerticalSpace = true;
		}
		String expand = e.getAttribute("expand");
		if (expand.equals("both") || expand.equals("horizontal")) {
			gd.horizontalAlignment = SWT.FILL;
		}
		if (expand.equals("both") || expand.equals("vertical")) {
			gd.verticalAlignment = SWT.FILL;
		}
		try {
			gd.widthHint = Integer.parseInt(e.getAttribute("width"));
		} catch (NumberFormatException ex) {
			// ignore
		}
		try {
			gd.heightHint = Integer.parseInt(e.getAttribute("height"));
		} catch (NumberFormatException ex) {
			// ignore
		}
		try {
			gd.horizontalSpan = Integer.parseInt(e.getAttribute("colspan"));
		} catch (NumberFormatException ex) {
			// ignore
		}
		try {
			gd.verticalSpan = Integer.parseInt(e.getAttribute("rowspan"));
		} catch (NumberFormatException ex) {
			// ignore
		}
		return gd;
	}
	
	private void refreshColor(String prefid) {
		
	}
	
	private void refreshMetaColors() {
		for (String id : windows.keySet()) {
			StyledText st = windows.get(id);
			if (prefs.isDefault(WINDOW_PREFIX + id + FOREGROUND_SUFFIX)) {
				st.getForeground().dispose();
				st.setForeground(new Color(Display.getCurrent(), PreferenceConverter.getColor(prefs, PREF_FOREGROUND_COLOR)));
			}
			if (prefs.isDefault(WINDOW_PREFIX + id + BACKGROUND_SUFFIX)) {
				st.getBackground().dispose();
				st.setBackground(new Color(Display.getCurrent(), PreferenceConverter.getColor(prefs, PREF_BACKGROUND_COLOR)));
			}
		}
		if (metabackground != null) {
			metabackground.dispose();
		}
		metabackground = new Color(Display.getCurrent(), PreferenceConverter.getColor(prefs, PREF_META_BACKGROUND_COLOR));
		for (Composite c : composites) {
			c.setBackground(metabackground);
		}
		for (Label l : labels.values()) {
			l.setBackground(metabackground);
		}
	}
	
	private void refreshFonts() {
		for (String id : windows.keySet()) {
			if (prefs.isDefault(WINDOW_PREFIX + id + FONT_SUFFIX)) {
				StyledText st = windows.get(id);
				st.getFont().dispose();
				st.setFont(new Font(Display.getCurrent(), PreferenceConverter.getFontData(prefs, PREF_MAIN_FONT)));
			}
		}
	}
	
	private void loadMacros(String characterID) {
		macros = MacroPreferencePage.getMacros(characterID);
	}
	
	private void loadHighlights(String characterID) {
		highlights = HighlightsPreferencePage.getHighlights(commandEntry.getDisplay(), characterID);
	}
	
	private void loadVariables(String characterID) {
		variables = new VariableService(characterID, VariablesPreferencePage.getVariables(characterID));
	}
	
	private void addHighlight(Highlight h) {
		highlights.put(h.getString(), h);
//		reHighlightAll();
	}
	
	private void removeHighlight(Highlight h) {
		Highlight oldHighlight = highlights.remove(h.getString());
		oldHighlight.getBackground().dispose();
		oldHighlight.getForeground().dispose();
//		reHighlightAll();
	}
	
	private void modifyHighlight(Highlight old, Highlight current) {
		if (old.getString().equals(current.getString())) {  // Only the color changes
			old = highlights.get(old.getString());
			if (old.getForeground() != current.getForeground()) {
				replaceForegroundColor(old.getForeground(), current.getForeground());
			}
			if (old.getBackground() != current.getBackground()) {
				replaceBackgroundColor(old.getBackground(), current.getBackground());
			}
			highlights.put(current.getString(), current);
		} else {   // The string itself changed
			Highlight oldHighlight = highlights.remove(old.getString());
			oldHighlight.getBackground().dispose();
			oldHighlight.getForeground().dispose();
			addHighlight(current);
		}
	}
	
	private void reHighlightAll() {
		gameView.setStyleRange(null);
		for (String preset : presetRanges.keySet()) {
			Color fg = presets.get(preset);
			for (StyleRange r : presetRanges.get(preset)) {
				StyleRange range = (StyleRange) r.clone();
				range.foreground = fg;
				gameView.setStyleRange(range);
			}
		}
		String text = gameView.getText();
		int i;
		for (String s : highlights.keySet()) {
			i = 0;
			Highlight h = highlights.get(s);
			while ((i = text.indexOf(s, i+1)) > -1) {
				gameView.setStyleRange(new StyleRange(i, s.length(), h.getForeground(), h.getBackground()));
			}
		}
	}
	
	private void loadPresets() {
		loadPreset("bold");
		loadPreset("speech");
		loadPreset("whisper");
		loadPreset("roomName");
	}
	
	private void loadPreset(String presetID) {
		Color oldColor = presets.get(presetID);
		Color newColor = new Color(Display.getCurrent(), PreferenceConverter.getColor(EclipsefePlugin.getDefault().getPreferenceStore(), PRESET_PREFIX + presetID));
		presets.put(presetID, newColor);
		replaceForegroundColor(oldColor, newColor);
		if (presetRanges.get(presetID) == null) {
			presetRanges.put(presetID, new ArrayList<StyleRange>());
		}
	}
	
	private void replaceForegroundColor(Color oldColor, Color newColor) {
		if ((oldColor != null) && (oldColor != newColor)) {
			StyleRange[] ranges = gameView.getStyleRanges();
			for (StyleRange r : ranges) {
				if (r.foreground == oldColor) {
					r.foreground = newColor;
					gameView.redraw();
				}
			}
			oldColor.dispose();
		}
	}
	
	private void replaceBackgroundColor(Color oldColor, Color newColor) {
		if ((oldColor != null) && (oldColor != newColor)) {
			StyleRange[] ranges = gameView.getStyleRanges();
			for (StyleRange r : ranges) {
				if (r.background == oldColor) {
					r.background = newColor;
				}
			}
			oldColor.dispose();
		}
	}
	
	private void sendCommands() {
		String text = commandEntry.getText();
		if (text.contains("\n")) {
			while (text.contains("\n")) {
				String command = text.substring(0, text.indexOf("\n"));
				sendCommand(command);
				text = text.substring(text.indexOf("\n") + 1);
			}
			commandEntry.setText(text);
			commandEntry.setSelection(commandEntry.getCharCount());
			currentCommand = -1;
		}
	}
	
	public void sendCommand(String command) {
		sendCommand(command, true);
	}
	
	public void disconnect() {
		displayError("Connection to game has been lost.");
		status = FEView.Status.DISCONNECTED;
		try {
			gameConnection.close();
		} catch (IOException e) {
			// Oh well
		}
	}
	
	public void sendCommand(String command, boolean addToHistory) {
		sendCommand(command, addToHistory, false);
	}
	
	public void sendHiddenCommand(String command) {
		boolean echo = setLocalEcho(false);
		sendCommand(command, false);
		setLocalEcho(echo);
	}
	
	public void sendCommand(String command, boolean addToHistory, boolean sendVerbatim) {
		if (Display.getCurrent() != null) {
			if ((command.length() > 0) && (command.charAt(command.length() - 1) == 13)) {
				command = command.substring(0, command.length() - 1);
			}
			if (!sendVerbatim) {
				for (ICommandProcessor cp : EclipsefePlugin.getCommandProcessors()) {
					command = cp.processRawCommand(command, this);
					if (command == null) {
						return;
					}
				}
				command = variables.variableReplacement(command);
				for (ICommandProcessor cp : EclipsefePlugin.getCommandProcessors()) {
					command = cp.processCommand(command, this);
					if (command == null) {
						return;
					}
				}
			}
			if (localecho) {
				prompt = false;
				appendText(command + '\n', false);
			}
			send(command + '\n');
			//output.print(command + '\n');
			if (addToHistory) {
				addToHistory(command);
			}
		} else {
			Display.getDefault().syncExec(new CommandRunnable(command, addToHistory, this));
		}
	}
	
	private void addToHistory(String command) {
		if (command.length() > 0) {
			sentCommands.add(0, command);
		}
	}
	
	public void fakeCommand(String command) {
		fakeCommand(command, true);
	}
	
	public void fakeCommand(String command, boolean addToHistory) {
		if (Display.getCurrent() != null) {
			if (localecho) {
				appendText(command + '\n', false);
			}
			if (addToHistory) {
				addToHistory(command);
			}
		} else {
			Display.getDefault().syncExec(new FakeCommandRunnable(command, addToHistory, this));
		}
	}
	
	public boolean setLocalEcho(boolean value) {
		boolean temp = localecho;
		localecho = value;
		return temp;
	}
	
	private void addText(String window, String text) {
		if (text.length() == 0) {
			return;
		}
		StyledText st = windows.get(window);
		if (st == null) {
			return;
		}
		if (st.isDisposed()) {
			return;
		}
		if (EclipsefePlugin.DEBUG) {
			System.out.println("Adding text: " + text);
		}
		ScrollBar bar = st.getVerticalBar();
		boolean atBottom = (bar.getMaximum() <= (bar.getSelection() + bar.getPageIncrement()));
		st.append(text);
		if (atBottom) {
			st.invokeAction(ST.TEXT_END);
		}
		if (!bar.isVisible() && (bar.getMaximum() > 1)) {
			bar.setVisible(true);
		} else if (bar.isVisible() && (bar.getMaximum() <= 1)) {
			bar.setVisible(false);
		}
	}
	
	public void appendText(String text) {
		appendText("main", text, true);
	}
	
	public void appendText(String text, boolean prefixNewline) {
		appendText("main", text, prefixNewline);
	}
	
	public void appendText(String window, String text, boolean prefixNewline) {
		if (Display.getCurrent() != null) {
			window = destWindow(window);
			if (window == null) {
				return;
			}
			if (prefixNewline && trailingNewlines.containsKey(window) && !trailingNewlines.get(window) && playmode) {
				text = '\n' + text;
			}
			addText(window, text);
			if (text.endsWith("\n") && (text.length() > 1)) {
				trailingNewlines.put(window, true);
			} else {
				trailingNewlines.put(window, false);
			}
		} else {
			Display.getDefault().syncExec(new TextAdderRunnable(window, text, this, prefixNewline));
		}
	}
	
	public VariableService getVariableService() {
		return variables;
	}
	
	public void appendText(String text, List<StyleRange> styles) {
		appendText("main", text, styles, true);
	}
	
	public void appendText(String window, String text, List<StyleRange> styles) {
		appendText(window, text, styles, true);
	}
	
	public void appendText(String text, List<StyleRange> styles, boolean prefixNewline) {
		appendText("main", text, styles, prefixNewline);
	}
	
	public void appendText(String window, String text, List<StyleRange> styles, boolean prefixNewline) {
		if (Display.getCurrent() != null) {
			window = destWindow(window);
			if (window == null) {
				return;
			}
			if (text.endsWith("\n") && playmode) {
				int len = text.length();
				text = text.substring(0, len - 1);
				for (StyleRange s : styles) {
					if (s.start + s.length == len) {
						s.length--;
					}
				}
			}
			StyledText st = windows.get(window);
			if (!st.isDisposed()) {
				int offset = st.getCharCount();
				if (prefixNewline && trailingNewlines.containsKey(window) && !trailingNewlines.get(window) && (st.getCharCount() != 0) && playmode) {
					text = '\n' + text;
					offset++;
				}
				if ((Boolean) st.getData("timestamps")) {
					Calendar cal = Calendar.getInstance();
					int hour = cal.get(Calendar.HOUR);
					if (hour == 0) {
						hour = 12;
					}
					int minute = cal.get(Calendar.MINUTE);
					int ampm = cal.get(Calendar.AM_PM);
					text = text + " [" + hour + ":" + (minute < 10 ? "0" : "") + minute + " " + (ampm == Calendar.AM ? "AM" : "PM") + "]";
				}
				appendText(window, text, false);
				for (StyleRange s : styles) {
					s.start += offset;
					st.setStyleRange(s);
				}
				adjustText(window);
			}
		} else {
			Display.getDefault().syncExec(new TextAdderRunnable(window, text, styles, this, prefixNewline));
		}
	}
	
	private void adjustText(String window) {
		int maxLength = prefs.getInt(PREF_MAX_SCROLLBACK);
		if (maxLength > 0) {
			StyledText st = windows.get(window);
			if (st == null) {
				return;
			}
			if (st.getLineCount() > maxLength) {
				int linesToRemove = st.getLineCount() - maxLength + (maxLength / 10);
				int end = st.getOffsetAtLine(linesToRemove);
				st.replaceTextRange(0, end, "");
			}
		}
	}
		
	public void setStatus(String status, String text, int percentage) {
		if (Display.getCurrent() != null) {
			Label l = progressBars.get(status);
			if (l != null) {
				l.setData("percent", percentage / 100.0);
				l.setText(text);
				l.redraw();
			}
		} else {
			Display.getDefault().syncExec(new StatusUpdateRunnable(this, status, text, percentage));
		}
	}
	
	public void registerWindow(String id, String ifClosed, String closedStyle) {
		if (ifClosed == null) {
			closedWindowRouting.put(id, "main");
		} else {
			closedWindowRouting.put(id, ifClosed);
		}
		if (EclipsefePlugin.DEBUG) {
			System.out.println("Window being registered: " + id + "; ifclosed=" + ifClosed + "; closedStyle=" + closedStyle);
		}
	}
	
	public String destWindow(String window) {
		StyledText st = windows.get(window);
		while (st == null) {
			window = closedWindowRouting.get(window);
			if (window == null) {
				window = "main";
			} else if (window.equals("")) {
				return null;
			}
			st = windows.get(window);
		}
		return window;
	}
	
	public void clearWindow(String window) {
		if (Display.getCurrent() != null) {
			StyledText st = windows.get(window);
			if (st != null) {
				st.setText("");
				st.setStyleRanges(new StyleRange[0]);
			}
		} else {
			Display.getDefault().syncExec(new ClearWindowRunnable(window, this));
		}
	}
	
	public void setGameInfo(String name, String game) {
		if (Display.getCurrent() != null) {
			this.setPartName(game + ": " + name);
		} else {
			Display.getDefault().syncExec(new GameInfoRunnable(name, game, this));
		}
	}
	
	public void displayError(String msg) {
		appendText(" *\n * " + msg + "\n *\n");
	}
	
	private String remainder = "";
	
	void incomingData(byte[] data, int len) {
		if (gameView.isDisposed()) {
			return;
		}
		for (StyledText st : windows.values()) {
			st.setRedraw(false);
		}
		remainder = remainder + new String(data, 0, len);
		String[] lines = remainder.split("(\\r)?\\n", -1);
		remainder = lines[lines.length - 1];
		LineAdder exec = null;
		for (int i = 0; i < lines.length - 1; i++) {
			String line = lines[i];
			//if (!playmode) {
				line = line + '\n';
			//}
			if (status == Status.CONNECTING) {
				if (line.equals("Please wait for connection to game server.\n")) {
					send("\n");
					status = Status.CONNECTED;
					List<IServiceProvider> providers = EclipsefePlugin.getServiceProviders();
					for (IServiceProvider p : providers) {
						p.startup(this);
					}
				}
			}
			exec = setLine(line);
			if (exec != null) {
				exec.run();
			}
//			if (exec == null) {
//				exec = setLine(line);
//				newExec = null;
//			} else {
//				newExec = setLine(line);
//				if (newExec != null) {
//					if (newExec.getWindow().equals(exec.getWindow())) {
//						exec.add(newExec);
//						newExec = null;
//					} else {
//						exec.setImmediate(true);
//					}
//				}
//			}
//			if ((exec != null) && exec.getImmediate()) {
//				if (exec != null) {
//					exec.run();
//					exec = newExec;
//					newExec = null;
//					if (((exec != null) && exec.getImmediate())) {
//						exec.run();
//						exec = null;
//					}
//				}
//			}
		}
		if (!playmode) {
			exec = setLine(remainder);
			if (exec != null) {
				exec.run();
			}
			remainder = "";
		}
		for (StyledText st : windows.values()) {
			st.setRedraw(true);
			st.redraw();
		}
	}
	
	
	public LineAdder setLine(String line) {
		streamThisLine = stream;
		ArrayList<StyleRange> styles = new ArrayList<StyleRange>();
		if (line.length() > 0) {
			for (IInputProcessor ip : EclipsefePlugin.getInputProcessors()) {
				line = ip.processRawString(line, this);
			}
			if (line.length() == 0) {
				return null;
			}
			line = processIncomingLine(line, styles);
			line = processRawText(line, styles);
			String dest = destWindow(streamThisLine);
			if ((line.length() > 0) && !addedPrompt && (dest != null) && dest.equals("main")) {
				prompt = false;
			} else if (line.length() == 0) {
				return null;
			}
			for (String preset : presets.keySet()) {
				Color c = presets.get(preset);
				for (StyleRange r : styles) {
					if (r.foreground == c) {
						StyleRange range = (StyleRange) r.clone();
						range.foreground = null;
						range.background = null;
						presetRanges.get(preset).add(range);
					}
				}
			}
		}
		addedPrompt = false;
		return new LineAdder(streamThisLine, line, styles, this);
	}
	
//		public void run() {
//			if (!gameView.isDisposed()) {
//				ArrayList<StyleRange> styles = new ArrayList<StyleRange>();
//				line = processIncomingLine(styles, gameView.getCharCount());
//				line = processRawText(line);
//				addText(line, true);
//				for (StyleRange style : styles) {
//					gameView.setStyleRange(style);
//				}
//			}
//		}
	
	private String processRawText(String line, ArrayList<StyleRange> styles) {
		for (IInputProcessor ip : EclipsefePlugin.getInputProcessors()) {
			line = ip.processParsedString(line, streamThisLine, styles, FEView.this);
		}
		int i;
		String lowerLine = line.toLowerCase();
		for (String s : highlights.keySet()) {
			i = -1;
			Highlight h = highlights.get(s);
			if (h.isEntireLine() & h.isIgnoreCase()) {
				s = s.toLowerCase();
				if (lowerLine.indexOf(s) > -1) {
					styles.add(new StyleRange(0, line.length(), h.getForeground(), h.getBackground()));
				}
			} else if (h.isEntireLine()) {
				if (line.indexOf(s) > -1) {
					styles.add(new StyleRange(0, line.length(), h.getForeground(), h.getBackground()));
				}
			} else if (h.isIgnoreCase()) {
				s = s.toLowerCase();
				while ((i = lowerLine.indexOf(s, i+1)) > -1) {
					styles.add(new StyleRange(i, s.length(), h.getForeground(), h.getBackground()));
				}
			} else {
				while ((i = line.indexOf(s, i+1)) > -1) {
					styles.add(new StyleRange(i, s.length(), h.getForeground(), h.getBackground()));
				}
			}
		}
		return line;
	}
	
	private String processIncomingLine(String line, ArrayList<StyleRange> ranges) {
		char[] lineArray = line.toCharArray();
		StringBuilder results = new StringBuilder();
		int inputIndex = 0;
		int outputIndex = 0;
		for ( ; inputIndex < lineArray.length; inputIndex++) {
			if (lineArray[inputIndex] == '<') {
				int newIndex = processTag(lineArray, inputIndex, outputIndex, ranges);
				if (inputIndex != newIndex) {
					inputIndex = newIndex;
					continue;
				}
			}
			if (lineArray[inputIndex] == '&') {
				if (playmode) {
					int previousLength = results.length();
					inputIndex += interpretEntity(lineArray, inputIndex, results) - 1;
					outputIndex += results.length() - previousLength;
					continue;
				}
			}
			outputIndex += addChar(lineArray[inputIndex], results);
		}
		if (bold > 0) {
			StyleRange range = new StyleRange(boldstart, outputIndex - boldstart, presets.get("bold"), null, SWT.NONE);
			ranges.add(range);
			boldstart = 0;
		}
		if (currentPreset != null) {
			Color color = presets.get(currentPreset);
			if (color != null) {
				StyleRange range = new StyleRange(presetstart, outputIndex - presetstart, color, null, SWT.NONE);
				ranges.add(range);
			}
			presetstart = 0;
		}
		return results.toString();
	}
	
	private int processTag(char[] input, int i, int o, ArrayList<StyleRange> ranges) {
		if (!playmode) {
			if ((input.length >= i + 5) && new String(input, i, 5).equals("<mode")) {
				if (EclipsefePlugin.DEBUG) {
					System.out.println("I found a mode: " + new String(input, i, input.length - i));
				}
				if (new String(input, i+6, 11).equals("id=\"GAME\"/>")) {
					playmode = true;
				}
			}
		}
		if (playmode) {
			StringBuilder tagbuf = new StringBuilder();
			while (input[i] != '>') {
				tagbuf.append(input[i]);
				i++;
			}
			tagbuf.append(input[i]);
			String tag = tagbuf.toString();
			if (EclipsefePlugin.DEBUG) {
				System.out.println("Found tag: " + tag);
			}
			if (tag.equals("<pushBold/>")) {
				pushBold(o, ranges);
			} else if (tag.equals("<popBold/>")) {
				popBold(o, ranges);
				return i;
			} else if (tag.equals("<b>")) {
				pushBold(o, ranges);
			} else if (tag.equals("</b>")) {
				popBold(o, ranges);
			} else if (tag.startsWith("<preset id=")) {
				currentPreset = extractAttribute(tag, "id");
				presetstart = o;
			} else if (tag.equals("</preset>")) {
				Color color = presets.get(currentPreset);
				if (color == null) {
					System.out.println("Found an unrecognized preset: " + currentPreset);
				} else {
					if (o > presetstart) {
						StyleRange range = new StyleRange(presetstart, o - presetstart, color, null, SWT.NONE);
						ranges.add(range);
					}
				}
				currentPreset = null;
			} else if (tag.startsWith("<style id=")) {
				String id = extractAttribute(tag, "id");
				if (id.equals("")) {
					Color color = presets.get(currentPreset);
					if (color == null) {
						System.out.println("Found an unrecognized preset: " + currentPreset);
					} else {
						if (o > presetstart) {
							StyleRange range = new StyleRange(presetstart, o - presetstart, color, null, SWT.NONE);
							ranges.add(range);
						}
					}
					currentPreset = null;
				} else {
					currentPreset = id;
					presetstart = o;
				}
				return i;
			} else if (tag.startsWith("<mode id=")) {
				if (!tag.equals("<mode id=\"GAME\"/>")) {
					playmode = false;
				}
			} else if (tag.startsWith("<pushStream ")) {
				pushStream(tag);
				return i;
			} else if (tag.startsWith("<popStream")) {
				popStream(tag);
			} else if (tag.startsWith("<clearStream")) {
				clearWindow(extractAttribute(tag, "id"));
			} else if (tag.startsWith("<compDef ")) {
				String id = extractAttribute(tag, "id");
				StyledText st = windows.get(stream);
				if (st != null) {
					st.setData("compdef " + id + " start", o + st.getCharCount());
					currentCompdef = id;
				}
			} else if (tag.startsWith("</compDef ")) {
				if (currentCompdef != null) {
					StyledText st = windows.get(stream);
					if (st != null) {
						st.setData("compdef " + currentCompdef + " end", o + st.getCharCount());
						currentCompdef = null;
					}
				}
			} else if (tag.startsWith("<component ")) {
				inComponent = true;
				componentBuilder.setLength(0);
			} else if (tag.equals("</component>")) {
				inComponent = false;
			} else if (tag.startsWith("<streamWindow ")) {
				String id = extractAttribute(tag, "id");
				String ifClosed = extractAttribute(tag, "ifClosed");
				String closedStyle = extractAttribute(tag, "styleIfClosed");
				registerWindow(id, ifClosed, closedStyle);
			} else if (tag.startsWith("<progressBar ")) {
				int v = Integer.parseInt(extractAttribute(tag, "value"));
				String text = extractAttribute(tag, "text");
				String statusId = extractAttribute(tag, "id");
				if (EclipsefePlugin.DEBUG) {
					System.out.println("Got a value = " + v + ", text was " + text + ", id was " + statusId + ", tag was " + tag);
				}
				setStatus(statusId, text, v);
			} else if (tag.startsWith("<label ")) {
				Label l = labels.get(extractAttribute(tag, "id"));
				if (l != null) {
					l.setText(extractAttribute(tag, "value"));
				}
			} else if (tag.startsWith("<left ") || tag.equals("<left>")) {
				inLeft = true;
			} else if (tag.equals("</left>")) {
				inLeft = false;
			} else if (tag.startsWith("<right ") || tag.equals("<right>")) {
				inRight = true;
			} else if (tag.equals("</right>")) {
				inRight = false;
			} else if (tag.startsWith("<spell ") || tag.equals("<spell>")) {
				inSpell = true;
			} else if (tag.equals("</spell>")) {
				inSpell = false;
			} else if (tag.startsWith("<app ")) {
				setGameInfo(extractAttribute(tag, "char"), extractAttribute(tag, "game"));
			} else if (tag.startsWith("<stream ")) {
				pushStream(tag);
			} else if (tag.equals("</stream>")) {
				stream = "main";
			} else if (tag.startsWith("<inv ")) {
				inInv = true;
			} else if (tag.equals("</inv>")) {
				inInv = false;
			} else if (tag.equals("</a>")) {
				return i;
			} else if (tag.equals("</d>")) {
				return i;
			} else if (tag.startsWith("<image ")) {
				String id = extractAttribute(tag, "id");
				if (id.equals("nsys") || id.equals("leftArm") || id.equals("rightArm")
						|| id.equals("rightLeg") || id.equals("leftLeg") || id.equals("head")
						|| id.equals("rightFoot") || id.equals("leftFoot") || id.equals("rightHand")
						|| id.equals("leftHand") || id.equals("rightEye") || id.equals("leftEye")
						|| id.equals("back") || id.equals("neck") || id.equals("chest")
						|| id.equals("abdomen")) {  // One of the injury locations
					if (injuries != null) {
						String name = extractAttribute(tag, "name");
						if (name.equals("Injury1")) {
							injuries.setData(id, 1);
						} else if (name.equals("Injury2")) {
							injuries.setData(id, 2);
						} else if (name.equals("Injury3")) {
							injuries.setData(id, 3);
						} else if (name.equals("Scar1")) {
							injuries.setData(id, -1);
						} else if (name.equals("Scar2")) {
							injuries.setData(id, -2);
						} else if (name.equals("Scar3")) {
							injuries.setData(id, -3);
						} else {
							injuries.setData(id, 0);
						}
						injuries.redraw();
					}
				}
			} else if (tag.startsWith("<radio ")) {
				String id = extractAttribute(tag, "id");
				try {
					int val = Integer.parseInt(extractAttribute(tag, "value"));
					if (injuries != null) {
						if (id.equals("injrRad")) {
							((Button) injuries.getData(id)).setSelection(val == 0 ? false : true);
						} else if (id.equals("scarRad")) {
							((Button) injuries.getData(id)).setSelection(val == 0 ? false : true);
						} else if (id.equals("bothRad")) {
							((Button) injuries.getData(id)).setSelection(val == 0 ? false : true);
						}
					}
				} catch (NumberFormatException e) {
					// just drop the tag
				}
			} else if (tag.startsWith("<roundTime ")) {
				try {
					int expiration = Integer.parseInt(extractAttribute(tag, "value"));
					if (progressBars.containsKey("roundtime")) {
						((RoundtimeUpdater) progressBars.get("roundtime").getData("updater")).setRoundtimeExpires(expiration);
					}
				} catch (NumberFormatException e) {
					// ignore it
				}
			} else if (tag.startsWith("<castTime ")) {
				try {
					int expiration = Integer.parseInt(extractAttribute(tag, "value"));
					if (progressBars.containsKey("castrt")) {
						((RoundtimeUpdater) progressBars.get("castrt").getData("updater")).setRoundtimeExpires(expiration);
					}
				} catch (NumberFormatException e) {
					// ignore it
				}
			} else if (tag.startsWith("<prompt ")) {
				try {
					int time = Integer.parseInt(extractAttribute(tag, "time"));
					timeOffset = (int) ((System.currentTimeMillis() / 1000) - time);
					if (progressBars.containsKey("roundtime")) {
						((RoundtimeUpdater) progressBars.get("roundtime").getData("updater")).setOffset(timeOffset);
					}
					if (progressBars.containsKey("castrt")) {
						((RoundtimeUpdater) progressBars.get("castrt").getData("updater")).setOffset(timeOffset);
					}
				} catch (NumberFormatException e) {
					// ignore it
				}
				StringBuilder buf = new StringBuilder();
				int j = i + 1;
				while (input[j] != '<') {
					if (input[j] == '&') {
						j += interpretEntity(input, j, buf);
					} else {
						buf.append(input[j]);
						j++;
					}
				}
				if (prompt) {
					i = j - 1;
				} else {
					prompt = true;
					addedPrompt = true;
				}
//					if (gameView.getTextRange(gameView.getCharCount() - buf.length(), buf.length()).equals(buf.toString())) {
//						i = j - 1;
//					} else {
//						addText("", true);
//					}
			} else if (tag.equals("</prompt>")) {
				// ignore
			} else {
				System.out.println("Tag dropped: " + tag);
			}
			if ((input.length > (i+1)) && (input[i+1] == '\n')) {
				i++;
			}
		}
		return i;
	}
	
	private String extractAttribute(String tag, String attribute) {
		char delim = '\''; 
		int start = tag.indexOf(" " + attribute + "=" + delim);
		if (start < 0) {
			delim = '"';
			start = tag.indexOf(" " + attribute + "=" + delim);
			if (start < 0) {
				return null;
			}
		}
		int end = tag.indexOf(Character.toString(delim), start + 3 + attribute.length());
		return tag.substring(start + 3 + attribute.length(), end);
	}
	
	private void pushBold(int o, ArrayList<StyleRange> ranges) {
		if (bold == 0) {
			boldstart = o;
		}
		bold++;
	}
	
	private void popBold(int o, ArrayList<StyleRange> ranges) {
		bold--;
		if (bold == 0) {
			if (o > boldstart) {
				StyleRange range = new StyleRange(boldstart, o - boldstart, presets.get("bold"), null, SWT.NONE);
				ranges.add(range);
			}
			boldstart = 0;
		} else if (bold < 0) {
			bold = 0;
		}
	}
	
	private void pushStream(String tag) {
		stream = extractAttribute(tag, "id");
		streamThisLine = stream;
		if (EclipsefePlugin.DEBUG) {
			System.out.println("Stream is now " + stream);
		}
	}
	
	private void popStream(String tag) {
		stream = "main";
		streamThisLine = stream;
	}
	
	private int addChar(char c, StringBuilder buf) {
		if (!inComponent && !inLeft && !inRight && !inSpell && !inInv) {
			buf.append(c);
			return 1;
		} else if (inComponent) {
			componentBuilder.append(c);
		}
		return 0;
	}
	
	private int interpretEntity(char[] input, int index, StringBuilder outputBuffer) {
		if (input[index] != '&') {
			return 0;
		}
		int originalIndex = index;
		index++;
		StringBuilder buf = new StringBuilder();
		while (Character.isLetterOrDigit(input[index])) {
			buf.append(input[index]);
			index++;
		}
		if (input[index] != ';') {
			index--;
		}
		if (entities.containsKey(buf.toString())) {
			addChar(entities.get(buf.toString()), outputBuffer);
		}
		return index - originalIndex + 1;
	}
	
	void send(String s) {
		try {
			int r = sc.write(ByteBuffer.wrap(s.getBytes()));
			if (r < s.length()) {
				sc.configureBlocking(true);
				s = s.substring(r);
				sc.write(ByteBuffer.wrap(s.getBytes()));
				sc.configureBlocking(false);
			}
		} catch (IOException e) {
			// do nothing
		}
	}
	
	public Shell getShell() {
		return commandEntry.getShell();
	}

	public void setFocus() {
		if (status.compareTo(Status.CONNECTING) >= 0) {
			commandEntry.setFocus();
		} else {
			loginScreen.setFocus();
		}
	}
	
	public void dispose() {
		try {
			if (gameConnection != null) {
				//output.println("exit");
				send("exit\n");
				gameConnection.close();
			}
		} catch (IOException e) {
			// do nothing
		}
		super.dispose();
		for (Color c : presets.values()) {
			c.dispose();
		}
	}
	
	private void connect(String hostname, int port, String key) {
		try {
			status = Status.CONNECTING;
			//gameConnection = new Socket(hostname, port);
			//listener = new SocketListener(gameConnection.getInputStream(), parentComposite.getDisplay());
			//new Thread(listener).start();
			InetSocketAddress addr = new InetSocketAddress(hostname, port);
			sc = SocketChannel.open(addr);
			gameConnection = sc.socket();
			new Thread(new SocketDispatcher(sc, this)).start();
			//output = new PrintStream(gameConnection.getOutputStream());
			//output.println("<c>" + key);
			//output.println("<c>/FE:WIZARD /VERSION:1.0.1.22 /P:WIN_XP /XML");
			send("<c>" + key + '\n');
			send("<c>/FE:ECLIPSEFE /VERSION:1.0.1.22 /P:WIN_XP /XML\n");
		} catch (UnknownHostException e) {
			gameView.setText("Unknown host.");
			status = Status.DISCONNECTED;
		} catch (IOException e) {
			gameView.setText("Network error.");
			status = Status.DISCONNECTED;
		}
	}

}
