/*
 * Created on Apr 23, 2006
 */
package com.simutronics.eclipsefe;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;

/**
 * @author Adam Vartanian
 */
public class InjuriesPaintListener implements PaintListener {

	private Control c;
	public static final String[] areas = new String[] {
		"nsys",
		"leftArm",
		"rightArm",
		"rightLeg",
		"leftLeg",
		"head",
		"rightFoot",
		"leftFoot",
		"rightHand",
		"leftHand",
		"rightEye",
		"leftEye",
		"back",
		"neck",
		"chest",
		"abdomen"
	};
	public static final Map<String, Point> areaLocations = new HashMap<String, Point>();
	
	static {
		areaLocations.put("nsys", new Point(8, 8));
		areaLocations.put("leftArm", new Point(36, 42));
		areaLocations.put("rightArm", new Point(16, 42));
		areaLocations.put("rightLeg", new Point(19, 65));
		areaLocations.put("leftLeg", new Point(33, 65));
		areaLocations.put("head", new Point(26, 24));
		areaLocations.put("rightFoot", new Point(1, 2));
		areaLocations.put("leftFoot", new Point(1, 2));
		areaLocations.put("rightHand", new Point(7, 42));
		areaLocations.put("leftHand", new Point(45, 42));
		areaLocations.put("rightEye", new Point(20, 18));
		areaLocations.put("leftEye", new Point(32, 18));
		areaLocations.put("back", new Point(44, 8));
		areaLocations.put("neck", new Point(26, 37));
		areaLocations.put("chest", new Point(26, 43));
		areaLocations.put("abdomen", new Point(26, 52));
	}
	
	public InjuriesPaintListener(Control c) {
		this.c = c;
		// these names are used because they're the names sent by SF as ids for
		//   the areas
		c.setData("nsys", 0);
		c.setData("leftArm", 0);
		c.setData("rightArm", 0);
		c.setData("rightLeg", 0);
		c.setData("leftLeg", 0);
		c.setData("head", 0);
		c.setData("rightFoot", 0);
		c.setData("leftFoot", 0);
		c.setData("rightHand", 0);
		c.setData("leftHand", 0);
		c.setData("rightEye", 0);
		c.setData("leftEye", 0);
		c.setData("back", 0);
		c.setData("neck", 0);
		c.setData("chest", 0);
		c.setData("abdomen", 0);
	}
	
	public void paintControl(PaintEvent e) {
		GC gc = e.gc;
		Image image = (Image) c.getData("image");
		Color bg = gc.getBackground();
		gc.drawImage(image, 0, 0);
		Color injuryColor = Display.getCurrent().getSystemColor(SWT.COLOR_RED);
		Color scarColor = Display.getCurrent().getSystemColor(SWT.COLOR_YELLOW);
		for (String loc : areas) {
			int value = (Integer) c.getData(loc);
			if (value > 0) {
				gc.setBackground(injuryColor);
			} else if (value < 0) {
				gc.setBackground(scarColor);
				value = -value;
			}
			if (value != 0) {
				Point p = areaLocations.get(loc);
				//gc.fillOval(8, 8, 4, 4);
				gc.fillOval(p.x - value - 1, p.y - value - 1, 2 + value * 2, 2 + value * 2);
			}
		}
		gc.setBackground(bg);
	}

}
