/*
 * Created on Apr 21, 2006
 */
package com.simutronics.eclipsefe;

import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;

/**
 * @author Adam Vartanian
 */
public class RoundtimeUpdater extends Thread {

	Control widget;
	private long offsetMillis = 0;
	long roundtimeExpiresMillis = 0;
	boolean running = false;
	private boolean shutdownRequested = false;
	
	public RoundtimeUpdater(Control target) {
		widget = target;
	}
	
	public void run() {
		while (!shutdownRequested) {
			try {
				if (!running) {
					sleep(500);
				} else {
					updateWidget();
					sleep(100);
				}
			} catch (InterruptedException e) {
				// do nothing, continue
			}
		}
	}
	
	public void setRoundtimeExpires(int time) {
		roundtimeExpiresMillis = ((long) time) * 1000 + offsetMillis;
		running = true;
		updateWidget();
	}
	
	public void setOffset(int offset) {
		if (offsetMillis != offset * 1000) {
			int offsetChange = (int) (offset - (offsetMillis / 1000));
			roundtimeExpiresMillis += offsetChange * 1000;
			offsetMillis = offset * 1000;
			updateWidget();
		}
	}
	
	public void shutdown() {
		shutdownRequested = true;
		interrupt();
	}
	
	private void updateWidget() {
		Display.getDefault().asyncExec(new WidgetUpdater(this));
	}
	
	private static class WidgetUpdater implements Runnable {
		
		private RoundtimeUpdater updater;
		
		public WidgetUpdater(RoundtimeUpdater upd) {
			updater = upd;
		}
		
		public void run() {
			int rt = (Integer) updater.widget.getData("roundtime");
			int newrt;
			if (updater.roundtimeExpiresMillis < System.currentTimeMillis()) {
				newrt = 0;
			} else {
				newrt = (int) ((updater.roundtimeExpiresMillis - System.currentTimeMillis()) / 1000 + 1);
			}
			if (rt != newrt) {
				updater.widget.setData("roundtime", newrt);
				updater.widget.redraw();
				updater.widget.update();
			}
			if (newrt <= 0) {
				updater.running = false;
			}
		}
	}
}
