/*
 * Created on Feb 1, 2006
 */
package com.simutronics.eclipsefe;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;

import org.eclipse.swt.widgets.Display;

/**
 * @author Adam Vartanian
 */
public class SocketDispatcher implements Runnable {

	private FEView view;
	private SocketChannel sc;
	private ByteBuffer buf = ByteBuffer.allocateDirect(4096);
	private byte[] buf2 = new byte[4096];
	private DispatchRunnable dispatcher;
	
	public SocketDispatcher(SocketChannel s, FEView v) throws IOException {
		view = v;
		sc = s;
		sc.configureBlocking(false);
		dispatcher = new DispatchRunnable(v);
	}

	public void run() {
		try {
			int read = 0;
			while (sc.isOpen() && (read != -1)) {
				buf.clear();
				read = sc.read(buf);
				if (read == 0) {
					sc.configureBlocking(true);
					read = sc.read(buf);
					sc.configureBlocking(false);
				}
				if (read > 0) {
					if (buf.hasArray()) {
						buf2 = buf.array();
					} else {
						buf.flip();
						buf.get(buf2, 0, read);
					}
					dispatcher.setData(buf2, read);
					Display.getDefault().syncExec(dispatcher);
				}
			}
		} catch (IOException e) {
			// we're done here
		}
		view.disconnect();
	}
	
	private static class DispatchRunnable implements Runnable {
		private byte[] buf;
		private int len;
		private FEView view;
		
		public DispatchRunnable(FEView v) {
			view = v;
		}
		
		public void setData(byte[] b, int l) {
			buf = b;
			len = l;
		}
		
		public void run() {
			view.incomingData(buf, len);
		}
	}

}
