/*
 * Created on May 29, 2005
 */
package com.simutronics.eclipsefe.preferences;

import org.eclipse.jface.preference.BooleanFieldEditor;
import org.eclipse.jface.preference.ColorFieldEditor;
import org.eclipse.jface.preference.DirectoryFieldEditor;
import org.eclipse.jface.preference.FieldEditorPreferencePage;
import org.eclipse.jface.preference.FontFieldEditor;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.IntegerFieldEditor;
import org.eclipse.jface.preference.PreferenceConverter;
import org.eclipse.jface.preference.StringFieldEditor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPreferencePage;

import com.simutronics.eclipsefe.EclipsefePlugin;

/**
 * @author Adam Vartanian
 */
public class FEPreferencePage extends FieldEditorPreferencePage implements IWorkbenchPreferencePage {

	public static final String PRESET_PREFIX = "color_preset_";
	
	public static final String PREF_FOREGROUND_COLOR = "defaultforegroundcolor";
	public static final String PREF_BACKGROUND_COLOR = "defaultbackgroundcolor";
	
	public static final String WINDOW_PREFIX = "window-";
	public static final String FOREGROUND_SUFFIX = "_foreground";
	public static final String BACKGROUND_SUFFIX = "_background";
	public static final String FONT_SUFFIX = "_font";
	public static final String TIMESTAMP_SUFFIX = "_timestamp";
	
	public static final String PREF_META_BACKGROUND_COLOR = "metabackgroundcolor";
	public static final String PREF_MAIN_FONT = "defaultfont";
	public static final String PREF_LOCALECHO = "localecho";
	public static final String PREF_PRESET_BOLD = PRESET_PREFIX + "bold";
	public static final String PREF_PRESET_SPEECH = PRESET_PREFIX + "speech";
	public static final String PREF_PRESET_WHISPER = PRESET_PREFIX + "whisper";
	public static final String PREF_PRESET_ROOMNAME = PRESET_PREFIX + "roomName";
	public static final String PREF_SCRIPT_PREFIX = "scriptprefix";
	public static final String PREF_SCRIPTS_DIRECTORY = "scriptsdirectory";
	public static final String PREF_MAX_SCROLLBACK = "maxscrollback";
	
	public static final RGB DEFAULT_FOREGROUND_COLOR = new RGB(128, 128, 128);
	public static final RGB DEFAULT_BACKGROUND_COLOR = new RGB(0, 0, 0);
	public static final RGB DEFAULT_META_BACKGROUND_COLOR = new RGB(0, 0, 0);
	public static final boolean DEFAULT_LOCALECHO = true;
	public static final RGB DEFAULT_PRESET_BOLD = new RGB(255, 255, 255);
	public static final RGB DEFAULT_PRESET_SPEECH = new RGB(128, 255, 128);
	public static final RGB DEFAULT_PRESET_WHISPER = new RGB(128, 255, 255);
	public static final RGB DEFAULT_PRESET_ROOMNAME = new RGB(128, 255, 255);
	public static final FontData DEFAULT_MAIN_FONT = new FontData("Courier New", 9, SWT.NORMAL);
	public static final String DEFAULT_SCRIPT_PREFIX = ".";
	public static final String DEFAULT_SCRIPTS_DIRECTORY = "";
	public static final int DEFAULT_MAX_SCROLLBACK = 2000;
	
	public FEPreferencePage() {
		super(FieldEditorPreferencePage.GRID);
		
		IPreferenceStore store = EclipsefePlugin.getDefault().getPreferenceStore();
		setPreferenceStore(store);
	}

	protected void createFieldEditors() {
		ColorFieldEditor foreground = new ColorFieldEditor(PREF_FOREGROUND_COLOR, "Default foreground color", getFieldEditorParent());
		addField(foreground);
		ColorFieldEditor background = new ColorFieldEditor(PREF_BACKGROUND_COLOR, "Default background color", getFieldEditorParent());
		addField(background);
		ColorFieldEditor metabackground = new ColorFieldEditor(PREF_META_BACKGROUND_COLOR, "Overall background color", getFieldEditorParent());
		addField(metabackground);
		ColorFieldEditor monsterbold = new ColorFieldEditor(PREF_PRESET_BOLD, "Monsterbold color", getFieldEditorParent());
		addField(monsterbold);
		ColorFieldEditor speech = new ColorFieldEditor(PREF_PRESET_SPEECH, "Speech color", getFieldEditorParent());
		addField(speech);
		ColorFieldEditor whispers = new ColorFieldEditor(PREF_PRESET_WHISPER, "Whisper color", getFieldEditorParent());
		addField(whispers);
		ColorFieldEditor roomName = new ColorFieldEditor(PREF_PRESET_ROOMNAME, "Room name color", getFieldEditorParent());
		addField(roomName);
		FontFieldEditor mainfont = new FontFieldEditor(PREF_MAIN_FONT, "Default font", "You swing a falchion at a rolton!", getFieldEditorParent());
		addField(mainfont);
		BooleanFieldEditor localecho = new BooleanFieldEditor(PREF_LOCALECHO, "Local Echo", getFieldEditorParent());
		addField(localecho);
		StringFieldEditor scriptprefix = new StringFieldEditor(PREF_SCRIPT_PREFIX, "Script Prefix", 5, getFieldEditorParent());
		addField(scriptprefix);
		DirectoryFieldEditor scriptdir = new DirectoryFieldEditor(PREF_SCRIPTS_DIRECTORY, "Scripts Location", getFieldEditorParent());
		addField(scriptdir);
		IntegerFieldEditor scrollback = new IntegerFieldEditor(PREF_MAX_SCROLLBACK, "Scrollback length", getFieldEditorParent());
		addField(scrollback);
	}
	
	public static void setDefaults(IPreferenceStore prefs) {
		PreferenceConverter.setDefault(prefs, FEPreferencePage.PREF_FOREGROUND_COLOR,
				FEPreferencePage.DEFAULT_FOREGROUND_COLOR);
		PreferenceConverter.setDefault(prefs, FEPreferencePage.PREF_BACKGROUND_COLOR,
				FEPreferencePage.DEFAULT_BACKGROUND_COLOR);
		PreferenceConverter.setDefault(prefs, FEPreferencePage.PREF_META_BACKGROUND_COLOR,
				FEPreferencePage.DEFAULT_META_BACKGROUND_COLOR);
		PreferenceConverter.setDefault(prefs, FEPreferencePage.PREF_MAIN_FONT,
				FEPreferencePage.DEFAULT_MAIN_FONT);
		PreferenceConverter.setDefault(prefs, FEPreferencePage.PREF_PRESET_BOLD,
				FEPreferencePage.DEFAULT_PRESET_BOLD);
		PreferenceConverter.setDefault(prefs, FEPreferencePage.PREF_PRESET_SPEECH,
				FEPreferencePage.DEFAULT_PRESET_SPEECH);
		PreferenceConverter.setDefault(prefs, FEPreferencePage.PREF_PRESET_WHISPER,
				FEPreferencePage.DEFAULT_PRESET_WHISPER);
		PreferenceConverter.setDefault(prefs, FEPreferencePage.PREF_PRESET_ROOMNAME,
				FEPreferencePage.DEFAULT_PRESET_ROOMNAME);
		prefs.setDefault(PREF_LOCALECHO, DEFAULT_LOCALECHO);
		prefs.setDefault(PREF_SCRIPT_PREFIX, DEFAULT_SCRIPT_PREFIX);
		prefs.setDefault(PREF_SCRIPTS_DIRECTORY, DEFAULT_SCRIPTS_DIRECTORY);
		prefs.setDefault(PREF_MAX_SCROLLBACK, DEFAULT_MAX_SCROLLBACK);
	}

	public void init(IWorkbench workbench) {

	}

}
