/*
 * Created on Oct 10, 2005
 */
package com.simutronics.eclipsefe.preferences;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;

import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.preference.ColorSelector;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.PreferenceConverter;
import org.eclipse.jface.preference.PreferencePage;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.StringConverter;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPreferencePage;

import com.simutronics.eclipsefe.EclipsefePlugin;
import com.simutronics.eclipsefe.Highlight;

import static com.simutronics.eclipsefe.preferences.FEPreferencePage.WINDOW_PREFIX;
import static com.simutronics.eclipsefe.preferences.FEPreferencePage.BACKGROUND_SUFFIX;
import static com.simutronics.eclipsefe.preferences.FEPreferencePage.FOREGROUND_SUFFIX;

/**
 * @author Adam Vartanian
 */
public class HighlightsPreferencePage extends PreferencePage implements IWorkbenchPreferencePage {

	public static final String PREF_HIGHLIGHTS = "highlights";
	public static final String DEFAULT_HIGHLIGHTS = "Global:\n";
	public static final String HIGHLIGHT_UPDATE_PREFIX = "highlight_update_";
	public static final String HIGHLIGHT_MODIFY = HIGHLIGHT_UPDATE_PREFIX + "modify";
	public static final String HIGHLIGHT_ADD = HIGHLIGHT_UPDATE_PREFIX + "add";
	public static final String HIGHLIGHT_REMOVE = HIGHLIGHT_UPDATE_PREFIX + "remove";
	public static final String HIGHLIGHT_SEPARATOR = "\f";
	
	private static class EditHighlightDialog extends Dialog {
		
		private String initialString;
		private Text stringBox;
		private ColorSelector foreground;
		private ColorSelector background;
		private String highlights;
		private Button ignoreCaseButton;
		private Button entireLineButton;
		private Button fgTransparentButton;
		private Button bgTransparentButton;
		String value;
		RGB fg;
		RGB bg;
		boolean ignoreCase;
		boolean entireLine;
		boolean fgTransparent;
		boolean bgTransparent;
		private Label errorLabel;
		private IPreferenceStore prefs;
		
		public EditHighlightDialog(Shell parent, String string, RGB fg, RGB bg, boolean entireLine, boolean ignoreCase, boolean fgTransparent, boolean bgTransparent, String characterHighlights, IPreferenceStore prefs) {
			super(parent);
			this.value = string;
			this.initialString = string;
			this.fg = fg;
			this.bg = bg;
			this.ignoreCase = ignoreCase;
			this.entireLine = entireLine;
			this.highlights = characterHighlights;
			this.prefs = prefs;
			this.fgTransparent = fgTransparent;
			this.bgTransparent = bgTransparent;
		}
		
		protected Control createDialogArea(Composite parent) {
			parent.getShell().setText("Edit Highlight");
			Composite p = (Composite) super.createDialogArea(parent);
			p.setLayout(new GridLayout(3, false));
			
			errorLabel = new Label(p, SWT.NONE);
			errorLabel.setForeground(parent.getDisplay().getSystemColor(SWT.COLOR_RED));
			GridData elGD = new GridData(SWT.CENTER, SWT.NONE, true, false);
			elGD.horizontalSpan = 3;
			elGD.widthHint = 150;
			errorLabel.setLayoutData(elGD);
			
			new Label(p, SWT.NONE).setText("Highlight String:");
			
			stringBox = new Text(p, SWT.BORDER | SWT.SINGLE);
			GridData kcGD = new GridData(SWT.NONE, SWT.NONE, false, false);
			kcGD.widthHint = 100;
			kcGD.horizontalSpan = 2;
			stringBox.setLayoutData(kcGD);
			stringBox.setText(value);
			stringBox.addModifyListener(new ModifyListener() {
				public void modifyText(ModifyEvent e) {
					String text = stringBox.getText();
					int i = highlights.indexOf(text + HIGHLIGHT_SEPARATOR);
					while ((i > 0) && (highlights.charAt(i-1) != '\t')) {
						i = highlights.indexOf(text + HIGHLIGHT_SEPARATOR, i+1);
					}
					if (!initialString.equals(text) && (i >= 0)) {
						errorLabel.setText("String already in use.");
						getButton(Dialog.OK).setEnabled(false);
					} else {
						errorLabel.setText("");
						getButton(Dialog.OK).setEnabled(true);
					}
				}
			});
			
			new Label(p, SWT.NONE).setText("Foreground:");
			
			foreground = new ColorSelector(p);
			if (fg != null) {
				foreground.setColorValue(fg);
			} else {
				foreground.setColorValue(PreferenceConverter.getColor(prefs, WINDOW_PREFIX + "main" + FOREGROUND_SUFFIX));
			}
			
			fgTransparentButton = new Button(p, SWT.CHECK);
			fgTransparentButton.setText("Transparent");
			if (fgTransparent) {
				fgTransparentButton.setSelection(true);
				foreground.setEnabled(false);
			} else {
				fgTransparentButton.setSelection(false);
			}
			fgTransparentButton.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent e) {
					foreground.setEnabled(!fgTransparentButton.getSelection());
				}
			});
			
			new Label(p, SWT.NONE).setText("Background:");
			
			background = new ColorSelector(p);
			if (bg != null) {
				background.setColorValue(bg);
			} else {
				background.setColorValue(PreferenceConverter.getColor(prefs, WINDOW_PREFIX + "main" + BACKGROUND_SUFFIX));
			}

			bgTransparentButton = new Button(p, SWT.CHECK);
			bgTransparentButton.setText("Transparent");
			if (bgTransparent) {
				bgTransparentButton.setSelection(true);
				background.setEnabled(false);
			} else {
				bgTransparentButton.setSelection(false);
			}
			bgTransparentButton.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent e) {
					background.setEnabled(!bgTransparentButton.getSelection());
				}
			});
			
			entireLineButton = new Button(p, SWT.CHECK);
			entireLineButton.setText("Entire Line");
			entireLineButton.setSelection(entireLine);
			
			ignoreCaseButton = new Button(p, SWT.CHECK);
			ignoreCaseButton.setText("Ignore Case");
			ignoreCaseButton.setSelection(ignoreCase);
			
			return p;
		}
		
		protected void okPressed() {
			value = stringBox.getText();
			fg = foreground.getColorValue();
			bg = background.getColorValue();
			ignoreCase = ignoreCaseButton.getSelection();
			entireLine = entireLineButton.getSelection();
			fgTransparent = fgTransparentButton.getSelection();
			if (fgTransparent) {
				fg = PreferenceConverter.getColor(prefs, WINDOW_PREFIX + "main" + FOREGROUND_SUFFIX);
			}
			bgTransparent = bgTransparentButton.getSelection();
			if (bgTransparent) {
				bg = PreferenceConverter.getColor(prefs, WINDOW_PREFIX + "main" + BACKGROUND_SUFFIX);
			}
			super.okPressed();
		}
	}
	
	Map<String, String> characters = new HashMap<String, String>();
	
	List characterList;
	Table highlightList;
	Button editButton;
	
	private ArrayList<PropertyChangeEvent> changes = new ArrayList<PropertyChangeEvent>();
	
	public HighlightsPreferencePage() {
		super();
	}

	public HighlightsPreferencePage(String title) {
		super(title);
	}

	public HighlightsPreferencePage(String title, ImageDescriptor image) {
		super(title, image);
	}

	protected Control createContents(Composite p) {
		noDefaultAndApplyButton();
		Composite parent = new Composite(p, SWT.NONE);
		parent.setLayout(new GridLayout(2, false));
		characterList = new List(parent, SWT.SINGLE | SWT.BORDER);
		GridData cListGD = new GridData(SWT.FILL, SWT.FILL, false, true);
		cListGD.widthHint = 100;
		characterList.setLayoutData(cListGD);
		characterList.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				updateHighlightList();
			}
		});
		
		Composite innerParent = new Composite(parent, SWT.NONE);
		innerParent.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		innerParent.setLayout(new GridLayout(3, false));
		
		highlightList = new Table(innerParent, SWT.SINGLE | SWT.FULL_SELECTION | SWT.BORDER);
		GridData highlightListGD = new GridData(SWT.FILL, SWT.FILL, true, true);
		highlightListGD.horizontalSpan = 3;
		highlightList.setLayoutData(highlightListGD);
		TableColumn firstColumn = new TableColumn(highlightList, SWT.NONE);
		firstColumn.setText("Highlight String");
		firstColumn.setWidth(250);
		highlightList.setHeaderVisible(true);
		highlightList.addSelectionListener(new SelectionAdapter(){
			public void widgetSelected(SelectionEvent e) {
				editButton.setEnabled(true);
			}
		});
		highlightList.setBackground(new Color(highlightList.getDisplay(), PreferenceConverter.getColor(doGetPreferenceStore(), WINDOW_PREFIX + "main" + BACKGROUND_SUFFIX)));
		
		Button addButton = new Button(innerParent, SWT.PUSH);
		addButton.setText("Add");
		addButton.setLayoutData(new GridData(SWT.FILL, SWT.NONE, true, false));
		addButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				EditHighlightDialog dialog = new EditHighlightDialog(getShell(), "", null, null, false, false, false, true, characters.get(characterList.getSelection()[0]), doGetPreferenceStore());
				dialog.open();
				if (dialog.getReturnCode() == Window.OK) {
					String fg = dialog.fgTransparent ? "null" : StringConverter.asString(dialog.fg);
					String bg = dialog.bgTransparent ? "null" : StringConverter.asString(dialog.bg);
					addHighlight(characterList.getSelection()[0], dialog.value + HIGHLIGHT_SEPARATOR + fg + HIGHLIGHT_SEPARATOR + bg + HIGHLIGHT_SEPARATOR + Boolean.toString(dialog.entireLine) + HIGHLIGHT_SEPARATOR + Boolean.toString(dialog.ignoreCase));
					int i = 0;
					while ((i < highlightList.getItemCount()) && (highlightList.getItem(i).getText(0).compareTo(dialog.value) < 0)) {
						i++;
					}
					TableItem newItem = new TableItem(highlightList, SWT.NONE, i);
					newItem.setText(dialog.value);
					newItem.setForeground(new Color(editButton.getDisplay(), dialog.fg));
					newItem.setBackground(new Color(editButton.getDisplay(), dialog.bg));
					newItem.setData("entireLine", dialog.entireLine);
					newItem.setData("ignoreCase", dialog.ignoreCase);
					newItem.setData("fgTransparent", dialog.fgTransparent);
					newItem.setData("bgTransparent", dialog.bgTransparent);
					changes.add(new PropertyChangeEvent(HighlightsPreferencePage.this, HIGHLIGHT_ADD, null, new Highlight(dialog.value, dialog.fgTransparent ? null : new Color(editButton.getDisplay(), dialog.fg), dialog.bgTransparent ? null : new Color(editButton.getDisplay(), dialog.bg), dialog.entireLine, dialog.ignoreCase)));
				}
			}
		});
		
		editButton = new Button(innerParent, SWT.PUSH);
		editButton.setText("Edit");
		editButton.setLayoutData(new GridData(SWT.FILL, SWT.NONE, true, false));
		editButton.setEnabled(false);
		editButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				if (highlightList.getSelectionCount() == 1) {
					int selected = highlightList.getSelectionIndex();
					TableItem toRemove = highlightList.getItem(selected);
					String character = characterList.getSelection()[0];
					EditHighlightDialog dialog = new EditHighlightDialog(getShell(), toRemove.getText(), toRemove.getForeground().getRGB(), toRemove.getBackground().getRGB(), (Boolean) toRemove.getData("entireLine"), (Boolean) toRemove.getData("ignoreCase"), (Boolean) toRemove.getData("fgTransparent"), (Boolean) toRemove.getData("bgTransparent"), characters.get(character), doGetPreferenceStore());
					dialog.open();
					if (dialog.getReturnCode() == Window.OK) {
						String fg = (Boolean) toRemove.getData("fgTransparent") ? "null" : StringConverter.asString(toRemove.getForeground().getRGB());
						String bg = (Boolean) toRemove.getData("bgTransparent") ? "null" : StringConverter.asString(toRemove.getBackground().getRGB());
						removeHighlight(character, toRemove.getText() + HIGHLIGHT_SEPARATOR + fg + HIGHLIGHT_SEPARATOR + bg + HIGHLIGHT_SEPARATOR + Boolean.toString((Boolean) toRemove.getData("entireLine")) + HIGHLIGHT_SEPARATOR + Boolean.toString((Boolean) toRemove.getData("ignoreCase")));
						fg = dialog.fgTransparent ? "null" : StringConverter.asString(dialog.fg);
						bg = dialog.bgTransparent ? "null" : StringConverter.asString(dialog.bg);
						addHighlight(character, dialog.value + HIGHLIGHT_SEPARATOR + fg + HIGHLIGHT_SEPARATOR + bg + HIGHLIGHT_SEPARATOR + Boolean.toString(dialog.entireLine) + HIGHLIGHT_SEPARATOR + Boolean.toString(dialog.ignoreCase));
						changes.add(new PropertyChangeEvent(HighlightsPreferencePage.this, HIGHLIGHT_MODIFY, new Highlight(toRemove.getText(), null, null, false, false), new Highlight(dialog.value, dialog.fgTransparent ? null : new Color(editButton.getDisplay(), dialog.fg), dialog.bgTransparent ? null : new Color(editButton.getDisplay(), dialog.bg), dialog.entireLine, dialog.ignoreCase)));
						toRemove.getForeground().dispose();
						toRemove.getBackground().dispose();
						toRemove.setText(dialog.value);
						toRemove.setForeground(new Color(editButton.getDisplay(), dialog.fg));
						toRemove.setBackground(new Color(editButton.getDisplay(), dialog.bg));
						toRemove.setData("entireLine", dialog.entireLine);
						toRemove.setData("ignoreCase", dialog.ignoreCase);
						toRemove.setData("fgTransparent", dialog.fgTransparent);
						toRemove.setData("bgTransparent", dialog.bgTransparent);
					}
				}
			}
		});
		
		Button removeButton = new Button(innerParent, SWT.PUSH);
		removeButton.setText("Remove");
		removeButton.setLayoutData(new GridData(SWT.FILL, SWT.NONE, true, false));
		removeButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				if (highlightList.getSelectionCount() == 1) {
					int selected = highlightList.getSelectionIndex();
					TableItem toRemove = highlightList.getItem(selected);
					String fg = (Boolean) toRemove.getData("fgTransparent") ? "null" : StringConverter.asString(toRemove.getForeground().getRGB());
					String bg = (Boolean) toRemove.getData("bgTransparent") ? "null" : StringConverter.asString(toRemove.getBackground().getRGB());
					String highlight = toRemove.getText() + HIGHLIGHT_SEPARATOR + fg + HIGHLIGHT_SEPARATOR + bg + HIGHLIGHT_SEPARATOR + Boolean.toString((Boolean) toRemove.getData("entireLine")) + HIGHLIGHT_SEPARATOR + Boolean.toString((Boolean) toRemove.getData("ignoreCase"));
					toRemove.getForeground().dispose();
					toRemove.getBackground().dispose();
					changes.add(new PropertyChangeEvent(HighlightsPreferencePage.this, HIGHLIGHT_REMOVE, new Highlight(toRemove.getText(), null, null, false, false), null));
					removeHighlight(characterList.getSelection()[0], highlight);
					highlightList.remove(selected);
					if (highlightList.getItemCount() > selected) {
						highlightList.setSelection(selected);
					} else {
						editButton.setEnabled(false);
					}
				}
			}
		});
		
		loadValues(doGetPreferenceStore().getString(PREF_HIGHLIGHTS));
		return parent;
	}

	public void init(IWorkbench workbench) {
		
	}
	
	void addHighlight(String character, String highlight) {
		String highlights = characters.get(character);
		if (highlights.length() == 0) {
			highlights = highlight;
		} else {
			highlights += "\t" + highlight;
		}
		characters.put(character, highlights);
		if (EclipsefePlugin.DEBUG) {
			System.out.println("Highlight added, highlights for " + character + " are now: " + highlights);
		}
	}
	
	void removeHighlight(String character, String highlight) {
		String highlights = characters.get(character);
		int i = highlights.indexOf(highlight);
		while ((i > 0) && (highlights.charAt(i-1) != '\t')) {
			i = highlights.indexOf(highlight, i+1);
		}
		if (i == 0) {
			if (highlights.length() == highlight.length()) {
				highlights = "";
			} else {
				highlights = highlights.substring(highlight.length()+1);
			}
		} else if (i > 0) {
			highlights = highlights.substring(0, i-1) + highlights.substring(i+highlight.length());
		}
		characters.put(character, highlights);
	}
	
	void updateHighlightList() {
		if (characterList.getSelectionCount() == 1) {
			for (TableItem t : highlightList.getItems()) {
				t.getBackground().dispose();
				t.getForeground().dispose();
			}
			highlightList.removeAll();
			String highlights = characters.get(characterList.getSelection()[0]);
			if (highlights.length() > 0) {
				IPreferenceStore prefs = EclipsefePlugin.getDefault().getPreferenceStore();
				String[] highlightArray = highlights.split("\t");
				Arrays.sort(highlightArray);
				for (String m : highlightArray) {
					TableItem item = new TableItem(highlightList, SWT.NONE);
					item.setText(m.substring(0, m.indexOf(HIGHLIGHT_SEPARATOR)));
					m = m.substring(m.indexOf(HIGHLIGHT_SEPARATOR) + 1);
					String foreground = m.substring(0, m.indexOf(HIGHLIGHT_SEPARATOR));
					m = m.substring(m.indexOf(HIGHLIGHT_SEPARATOR) + 1);
					String background = m.substring(0, m.indexOf(HIGHLIGHT_SEPARATOR));
					m = m.substring(m.indexOf(HIGHLIGHT_SEPARATOR) + 1);
					boolean entireLine = Boolean.parseBoolean(m.substring(0, m.indexOf(HIGHLIGHT_SEPARATOR)));
					m = m.substring(m.indexOf(HIGHLIGHT_SEPARATOR) + 1);
					boolean ignoreCase = Boolean.parseBoolean(m);
					if (foreground.equals("null")) {
						item.setForeground(new Color(highlightList.getDisplay(), PreferenceConverter.getColor(prefs, WINDOW_PREFIX + "main" + FOREGROUND_SUFFIX)));
					} else {
						item.setForeground(new Color(highlightList.getDisplay(), StringConverter.asRGB(foreground)));
					}
					if (background.equals("null")) {
						item.setBackground(new Color(highlightList.getDisplay(), PreferenceConverter.getColor(prefs, WINDOW_PREFIX + "main" + BACKGROUND_SUFFIX)));
					} else {
						item.setBackground(new Color(highlightList.getDisplay(), StringConverter.asRGB(background)));
					}
					
					item.setData("entireLine", entireLine);
					item.setData("ignoreCase", ignoreCase);
					item.setData("fgTransparent", foreground.equals("null"));
					item.setData("bgTransparent", background.equals("null"));
				}
				editButton.setEnabled(false);
			}
		}
	}
	
	private void loadValues(String pref) {
		characters.clear();
		characterList.removeAll();
		String[] chars = pref.split("\n");
		Arrays.sort(chars);
		for (String c : chars) {
			if (c.length() > 0) {
				String name = c.substring(0, c.indexOf(':'));
				characters.put(name, c.substring(c.indexOf(':') + 1));
				characterList.add(name);
			}
		}
		highlightList.clearAll();
	}
	
	public boolean performOk() {
		IPreferenceStore prefs = doGetPreferenceStore();
		StringBuilder results = new StringBuilder();
		for (String c : characters.keySet()) {
			results.append(c).append(':').append(characters.get(c)).append('\n');
		}
		prefs.setValue(PREF_HIGHLIGHTS, results.toString());
		for (PropertyChangeEvent e : changes) {
			prefs.firePropertyChangeEvent(e.getProperty(), e.getOldValue(), e.getNewValue());
		}
		changes.clear();
		return true;
	}
	
	protected void performDefaults() {
		super.performDefaults();
	}
	
	public static void setDefaults(IPreferenceStore prefs) {
		prefs.setDefault(PREF_HIGHLIGHTS, DEFAULT_HIGHLIGHTS);
	}
	
	protected IPreferenceStore doGetPreferenceStore() {
		return EclipsefePlugin.getDefault().getPreferenceStore();
	}
	
	public static Map<String, Highlight> getHighlights(Display d, String characterID) {
		IPreferenceStore prefs = EclipsefePlugin.getDefault().getPreferenceStore();
		String[] characters = prefs.getString(PREF_HIGHLIGHTS).split("\n");
		Map<String, Highlight> highlights = new TreeMap<String, Highlight>();
		String globalHighlights = null;
		String charHighlights = null;
		for (String character : characters) {
			if (character.length() > 0) {
				String name = character.substring(0, character.indexOf(':'));
				if (name.equals(characterID)) {
					charHighlights = character.substring(character.indexOf(':') + 1);
				} else if (name.equals(MacroPreferencePage.GLOBAL_CHARACTER)) {
					globalHighlights = character.substring(character.indexOf(':') + 1);
				}
			}
		}
		if ((globalHighlights != null) && (globalHighlights.length() > 0)) {
			addHighlights(d, globalHighlights.split("\t"), highlights);
		}
		if ((charHighlights != null) && (charHighlights.length() > 0)) {
			addHighlights(d, charHighlights.split("\t"), highlights);
		} else if (charHighlights == null) {
			prefs.setValue(PREF_HIGHLIGHTS, prefs.getString(PREF_HIGHLIGHTS) + characterID + ":\n");
		}
		return highlights;
	}
		
	private static void addHighlights(Display d, String[] highlightList, Map<String, Highlight> highlights) {
		for (String m : highlightList) {
			String string = m.substring(0, m.indexOf(HIGHLIGHT_SEPARATOR));
			m = m.substring(m.indexOf(HIGHLIGHT_SEPARATOR) + 1);
			String foreground = m.substring(0, m.indexOf(HIGHLIGHT_SEPARATOR));
			m = m.substring(m.indexOf(HIGHLIGHT_SEPARATOR) + 1);
			String background = m.substring(0, m.indexOf(HIGHLIGHT_SEPARATOR));
			m = m.substring(m.indexOf(HIGHLIGHT_SEPARATOR) + 1);
			boolean entireLine = Boolean.parseBoolean(m.substring(0, m.indexOf(HIGHLIGHT_SEPARATOR)));
			m = m.substring(m.indexOf(HIGHLIGHT_SEPARATOR) + 1);
			boolean ignoreCase = Boolean.parseBoolean(m);
			
			Color fg;
			if (foreground.equals("null")) {
				fg = null;
			} else {
				fg = new Color(d, StringConverter.asRGB(foreground));
			}
			Color bg;
			if (background.equals("null")) {
				bg = null;
			} else {
				bg = new Color(d, StringConverter.asRGB(background));
			}
			highlights.put(string, new Highlight(string, fg, bg, entireLine, ignoreCase));
		}
	}
	
	public void dispose() {
		for (TableItem t : highlightList.getItems()) {
			t.getForeground().dispose();
			t.getBackground().dispose();
		}
		highlightList.getBackground().dispose();
		super.dispose();
	}

}
