/*
 * Created on Oct 10, 2005
 */
package com.simutronics.eclipsefe.preferences;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.jface.bindings.keys.KeyStroke;
import org.eclipse.jface.bindings.keys.ParseException;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.PreferencePage;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPreferencePage;
import org.eclipse.ui.keys.IBindingService;

import com.simutronics.eclipsefe.EclipsefePlugin;

/**
 * @author Adam Vartanian
 */
public class MacroPreferencePage extends PreferencePage implements IWorkbenchPreferencePage {

	public static final String PREF_MACROS = "macros";
	public static final String GLOBAL_CHARACTER = "Global";
	public static final String BASE_MACRO_SET = "PAGE_UP {pageup}\tPAGE_DOWN {pagedown}\tM1+x {cut}\tM1+v {paste}\tM1+c {copy}\tARROW_UP {lastcommand}\tARROW_DOWN {nextcommand}\tM1+CR {sendlastcommand}\tM3+CR {sendsecondtolastcommand}\tM1+a {selectall}";
	public static final String DEFAULT_MACROS = GLOBAL_CHARACTER + ":" + BASE_MACRO_SET + "\n";
	
	private static class EditMacroDialog extends Dialog {
		
		private String initialKeystroke;
		private Text keyComboBox;
		private Text resultBox;
		private Label errorLabel;
		String keyStroke;
		String value;
		String macros;
		private IBindingService bindingService;
		
		public EditMacroDialog(Shell parent, String keyStroke, String value, String characterMacros) {
			super(parent);
			this.keyStroke = keyStroke;
			this.initialKeystroke = keyStroke;
			this.value = value;
			this.macros = characterMacros;
			this.bindingService = (IBindingService) EclipsefePlugin.getDefault().getWorkbench().getAdapter(IBindingService.class);
		}
		
		protected Control createDialogArea(Composite parent) {
			parent.getShell().setText("Edit Macro");
			Composite p = (Composite) super.createDialogArea(parent);
			p.setLayout(new GridLayout(2, false));
			
			errorLabel = new Label(p, SWT.NONE);
			errorLabel.setForeground(parent.getDisplay().getSystemColor(SWT.COLOR_RED));
			GridData elGD = new GridData(SWT.CENTER, SWT.NONE, true, false);
			elGD.horizontalSpan = 2;
			elGD.widthHint = 200;
			errorLabel.setLayoutData(elGD);
			
			new Label(p, SWT.NONE).setText("Key Combo:");
			
			keyComboBox = new Text(p, SWT.BORDER | SWT.SINGLE);
			GridData kcGD = new GridData(SWT.NONE, SWT.NONE, false, false);
			kcGD.widthHint = 150;
			keyComboBox.setLayoutData(kcGD);
			keyComboBox.addKeyListener(new KeyAdapter() {
				public void keyPressed(KeyEvent e) {
					String keyStroke = KeyStroke.getInstance(e.stateMask, e.keyCode).toString();
					keyComboBox.setText(keyStroke);
					e.doit = false;
					if (!MacroPreferencePage.validMacroKey(e)) {
						errorLabel.setText("Invalid macro key combination.");
						getButton(Dialog.OK).setEnabled(false);
					} else {
						int i = macros.indexOf(keyStroke);
						while ((i > 0) && (macros.charAt(i-1) != '\t')) {
							i = macros.indexOf(keyStroke, i+1);
						}
						if (!initialKeystroke.equals(keyStroke) && (i >= 0)) {
							errorLabel.setText("Key combination already in use.");
							getButton(Dialog.OK).setEnabled(false);
						} else {
							errorLabel.setText("");
							getButton(Dialog.OK).setEnabled(true);
						}
					}
				}
			});
			keyComboBox.addFocusListener(new FocusListener(){
				public void focusLost(FocusEvent e) {
					bindingService.setKeyFilterEnabled(true);
				}
				public void focusGained(FocusEvent e) {
					bindingService.setKeyFilterEnabled(false);
				}
			});
			keyComboBox.setText(keyStroke);
			
			new Label(p, SWT.NONE).setText("Result:");
			
			resultBox = new Text(p, SWT.BORDER | SWT.SINGLE);
			GridData resultGD = new GridData(SWT.FILL, SWT.NONE, true, false);
			resultGD.widthHint = IDialogConstants.ENTRY_FIELD_WIDTH;
			resultBox.setLayoutData(resultGD);
			resultBox.setText(value);
			
			return p;
		}
		
		protected void okPressed() {
			keyStroke = keyComboBox.getText();
			value = resultBox.getText();
			super.okPressed();
		}
	}
	
	Map<String, String> characters = new HashMap<String, String>();
	
	List characterList;
	Table macroList;
	Button editButton;
	
	public MacroPreferencePage() {
		super();
	}

	public MacroPreferencePage(String title) {
		super(title);
	}

	public MacroPreferencePage(String title, ImageDescriptor image) {
		super(title, image);
	}

	protected Control createContents(Composite p) {
		Composite parent = new Composite(p, SWT.NONE);
		parent.setLayout(new GridLayout(2, false));
		characterList = new List(parent, SWT.SINGLE | SWT.BORDER);
		GridData cListGD = new GridData(SWT.FILL, SWT.FILL, false, true);
		cListGD.widthHint = 100;
		characterList.setLayoutData(cListGD);
		characterList.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				updateMacroList();
			}
		});
		
		Composite innerParent = new Composite(parent, SWT.NONE);
		innerParent.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		innerParent.setLayout(new GridLayout(3, false));
		
		macroList = new Table(innerParent, SWT.SINGLE | SWT.FULL_SELECTION | SWT.BORDER);
		GridData macroListGD = new GridData(SWT.FILL, SWT.FILL, true, true);
		macroListGD.horizontalSpan = 3;
		macroList.setLayoutData(macroListGD);
		TableColumn firstColumn = new TableColumn(macroList, SWT.NONE);
		firstColumn.setText("Macro");
		firstColumn.setWidth(100);
		TableColumn secondColumn = new TableColumn(macroList, SWT.NONE);
		secondColumn.setText("Value");
		secondColumn.setWidth(150);
		macroList.setHeaderVisible(true);
		macroList.addSelectionListener(new SelectionAdapter(){
			public void widgetSelected(SelectionEvent e) {
				editButton.setEnabled(true);
			}
		});
		
		Button addButton = new Button(innerParent, SWT.PUSH);
		addButton.setText("Add");
		addButton.setLayoutData(new GridData(SWT.FILL, SWT.NONE, true, false));
		addButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				EditMacroDialog dialog = new EditMacroDialog(getShell(), "", "", characters.get(characterList.getSelection()[0]));
				dialog.open();
				if (dialog.getReturnCode() == Window.OK) {
					addMacro(characterList.getSelection()[0], dialog.keyStroke + " " + dialog.value);
					int i = 0;
					while ((i < macroList.getItemCount()) && (macroList.getItem(i).getText(0).compareTo(dialog.keyStroke) < 0)) {
						i++;
					}
					new TableItem(macroList, SWT.NONE, i).setText(new String[]{dialog.keyStroke, dialog.value});
				}
			}
		});
		
		editButton = new Button(innerParent, SWT.PUSH);
		editButton.setText("Edit");
		editButton.setLayoutData(new GridData(SWT.FILL, SWT.NONE, true, false));
		editButton.setEnabled(false);
		editButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				if (macroList.getSelectionCount() == 1) {
					int selected = macroList.getSelectionIndex();
					TableItem toRemove = macroList.getItem(selected);
					String macro = toRemove.getText(0) + " " + toRemove.getText(1);
					String character = characterList.getSelection()[0];
					EditMacroDialog dialog = new EditMacroDialog(getShell(), toRemove.getText(0), toRemove.getText(1),characters.get(character));
					dialog.open();
					if (dialog.getReturnCode() == Window.OK) {
						removeMacro(character, macro);
						addMacro(character, dialog.keyStroke + " " + dialog.value);
						toRemove.setText(0, dialog.keyStroke);
						toRemove.setText(1, dialog.value);
					}
				}
			}
		});
		
		Button removeButton = new Button(innerParent, SWT.PUSH);
		removeButton.setText("Remove");
		removeButton.setLayoutData(new GridData(SWT.FILL, SWT.NONE, true, false));
		removeButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				if (macroList.getSelectionCount() == 1) {
					int selected = macroList.getSelectionIndex();
					TableItem toRemove = macroList.getItem(selected);
					String macro = toRemove.getText(0) + " " + toRemove.getText(1);
					removeMacro(characterList.getSelection()[0], macro);
					macroList.remove(selected);
					if (macroList.getItemCount() > selected) {
						macroList.setSelection(selected);
					} else {
						editButton.setEnabled(false);
					}
				}
			}
		});
		
		loadValues(doGetPreferenceStore().getString(PREF_MACROS));
		return parent;
	}

	public void init(IWorkbench workbench) {
		
	}
	
	void addMacro(String character, String macro) {
		String macros = characters.get(character);
		if (macros.length() == 0) {
			macros = macro;
		} else {
			macros += "\t" + macro;
		}
		characters.put(character, macros);
		System.out.println("Macro added, macros for " + character + " are now: " + macros);
	}
	
	void removeMacro(String character, String macro) {
		String macros = characters.get(character);
		int i = macros.indexOf(macro);
		while ((i > 0) && (macros.charAt(i-1) != '\t')) {
			i = macros.indexOf(macro, i+1);
		}
		if (i == 0) {
			if (macros.length() == macro.length()) {
				macros = "";
			} else {
				macros = macros.substring(macro.length()+1);
			}
		} else if (i > 0) {
			macros = macros.substring(0, i-1) + macros.substring(i+macro.length());
		}
		characters.put(character, macros);
	}
	
	void updateMacroList() {
		if (characterList.getSelectionCount() == 1) {
			macroList.removeAll();
			String macros = characters.get(characterList.getSelection()[0]);
			if (macros.length() > 0) {
				String[] macroArray = macros.split("\t");
				Arrays.sort(macroArray);
				for (String m : macroArray) {
					try {
						String keyStroke = KeyStroke.getInstance(m.substring(0, m.indexOf(' '))).toString();
						new TableItem(macroList, SWT.NONE).setText(new String[]{keyStroke, m.substring(m.indexOf(' ') + 1)});
					} catch (ParseException e) {
						// ignore
					}
				}
				editButton.setEnabled(false);
			}
		}
	}
	
	private void loadValues(String pref) {
		characters.clear();
		characterList.removeAll();
		String[] chars = pref.split("\n");
		Arrays.sort(chars);
		for (String c : chars) {
			if (c.length() > 0) {
				String name = c.substring(0, c.indexOf(':'));
				characters.put(name, c.substring(c.indexOf(':') + 1));
				characterList.add(name);
			}
		}
		macroList.clearAll();
	}
	
	public boolean performOk() {
		IPreferenceStore prefs = doGetPreferenceStore();
		StringBuilder results = new StringBuilder();
		for (String c : characters.keySet()) {
			results.append(c).append(':').append(characters.get(c)).append('\n');
		}
		prefs.setValue(PREF_MACROS, results.toString());
		return true;
	}
	
	protected void performDefaults() {
		if (characterList.getSelectionCount() == 1) {
			characters.put(characterList.getSelection()[0], BASE_MACRO_SET);
			updateMacroList();
		}
		super.performDefaults();
	}
	
	public static void setDefaults(IPreferenceStore prefs) {
		prefs.setDefault(PREF_MACROS, DEFAULT_MACROS);
	}
	
	protected IPreferenceStore doGetPreferenceStore() {
		return EclipsefePlugin.getDefault().getPreferenceStore();
	}
	
	public static boolean validMacroKey(KeyEvent e) {
		return validMacroKey(e.keyCode, e.stateMask);
	}
	
	public static boolean validMacroKey(int keyCode, int stateMask) {
		if ((keyCode == SWT.CR) && (stateMask == SWT.NONE)) {
			return false;   // plain carraige return
		} else if ((0x20 <= keyCode) && (keyCode <= 0x7E) && ((stateMask == SWT.NONE) || (stateMask == SWT.SHIFT))) {
			return false;   // upper or lowercase letter
		} else if ((keyCode == SWT.ARROW_LEFT) || (keyCode == SWT.ARROW_RIGHT) || (keyCode == SWT.HOME) || (keyCode == SWT.END)) {
			return false;   // left/right arrow or home/end
		} else if (((keyCode == SWT.BS) || (keyCode == SWT.DEL)) && (stateMask == SWT.NONE)) {
			return false;   // backspace and del with no modifier
		}
		return true;
	}
	
	public static Map<KeyStroke, String> getMacros(String characterID) {
		IPreferenceStore prefs = EclipsefePlugin.getDefault().getPreferenceStore();
		String[] characters = prefs.getString(PREF_MACROS).split("\n");
		Map<KeyStroke, String> macros = new HashMap<KeyStroke, String>();
		String globalMacros = null;
		String charMacros = null;
		for (String character : characters) {
			if (character.length() > 0) {
				String name = character.substring(0, character.indexOf(':'));
				if (name.equals(characterID)) {
					charMacros = character.substring(character.indexOf(':') + 1);
				} else if (name.equals(GLOBAL_CHARACTER)) {
					globalMacros = character.substring(character.indexOf(':') + 1);
				}
			}
		}
		if ((globalMacros != null) && (globalMacros.length() > 0)) {
			addMacros(globalMacros.split("\t"), macros);
		}
		if ((charMacros != null) && (charMacros.length() > 0)) {
			addMacros(charMacros.split("\t"), macros);
		} else if (charMacros == null) {
			prefs.setValue(PREF_MACROS, prefs.getString(PREF_MACROS) + characterID + ":\n");
		}
		return macros;
	}
	
	private static void addMacros(String[] macrolist, Map<KeyStroke, String> macros) {
		for (String m : macrolist) {
			String keys = m.substring(0, m.indexOf(' '));
			String value = m.substring(m.indexOf(' ') + 1);
			try {
				macros.put(KeyStroke.getInstance(keys), value);
			} catch (ParseException e) {
				// ignore
			}
		}
	}

}
