/*
 * Created on Oct 9, 2005
 */
package com.simutronics.eclipsefe.preferences;

import org.eclipse.jface.preference.BooleanFieldEditor;
import org.eclipse.jface.preference.FieldEditorPreferencePage;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.StringFieldEditor;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPreferencePage;

import com.simutronics.eclipsefe.EclipsefePlugin;

/**
 * @author Adam Vartanian
 */
public class StoredInfoPreferencePage extends FieldEditorPreferencePage implements IWorkbenchPreferencePage {

	public static final String PREF_SAVED_ACCOUNT = "account_name";
	public static final String PREF_SAVED_INSTANCE = "saved_instance";
	public static final String PREF_SAVED_CHARACTER = "saved_character";
	public static final String PREF_AUTOSAVE_LOGIN_DATA = "autosave_login_data";

	public static final String DEFAULT_SAVED_ACCOUNT = "";
	public static final String DEFAULT_SAVED_INSTANCE = "";
	public static final String DEFAULT_SAVED_CHARACTER = "";
	public static final boolean DEFAULT_AUTOSAVE_LOGIN_DATA = true;
	
	public StoredInfoPreferencePage() {
		super(FieldEditorPreferencePage.GRID);
		
		IPreferenceStore store = EclipsefePlugin.getDefault().getPreferenceStore();
		setPreferenceStore(store);
	}

	protected void createFieldEditors() {
		StringFieldEditor account = new StringFieldEditor(PREF_SAVED_ACCOUNT, "Default account name:", getFieldEditorParent());
		addField(account);
		StringFieldEditor instance = new StringFieldEditor(PREF_SAVED_INSTANCE, "Default instance name:", getFieldEditorParent());
		addField(instance);
		StringFieldEditor character = new StringFieldEditor(PREF_SAVED_CHARACTER, "Default character name:", getFieldEditorParent());
		addField(character);
		BooleanFieldEditor autosave = new BooleanFieldEditor(PREF_AUTOSAVE_LOGIN_DATA, "Change on login", getFieldEditorParent());
		addField(autosave);
	}
	
	public static void setDefaults(IPreferenceStore prefs) {
		prefs.setDefault(StoredInfoPreferencePage.PREF_AUTOSAVE_LOGIN_DATA,
				StoredInfoPreferencePage.DEFAULT_AUTOSAVE_LOGIN_DATA);
		prefs.setDefault(StoredInfoPreferencePage.PREF_SAVED_ACCOUNT,
				StoredInfoPreferencePage.DEFAULT_SAVED_ACCOUNT);
		prefs.setDefault(StoredInfoPreferencePage.PREF_SAVED_CHARACTER,
				StoredInfoPreferencePage.DEFAULT_SAVED_CHARACTER);
		prefs.setDefault(StoredInfoPreferencePage.PREF_SAVED_INSTANCE,
				StoredInfoPreferencePage.DEFAULT_SAVED_INSTANCE);
	}

	public void init(IWorkbench workbench) {

	}

}
