/*
 * Created on Oct 10, 2005
 */
package com.simutronics.eclipsefe.preferences;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.PreferencePage;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPreferencePage;

import com.simutronics.eclipsefe.EclipsefePlugin;

/**
 * @author Adam Vartanian
 */
public class VariablesPreferencePage extends PreferencePage implements IWorkbenchPreferencePage {

	public static final String PREF_VARIABLES = "variables";
	public static final String DEFAULT_VARIABLES = "Global:\n";
	public static final String VARIABLE_SEPARATOR = "\f";
	
	private static class EditVariableDialog extends Dialog {
		
		private String initialName;
		private Text nameBox;
		private Text valueBox;
		private String variables;
		String name;
		String value;
		private Label errorLabel;
		
		public EditVariableDialog(Shell parent, String name, String value, String characterVars) {
			super(parent);
			this.name = name;
			this.initialName = name;
			this.value = value;
			this.variables = characterVars;
		}
		
		protected Control createDialogArea(Composite parent) {
			parent.getShell().setText("Edit Variable");
			Composite p = (Composite) super.createDialogArea(parent);
			p.setLayout(new GridLayout(2, false));
			
			errorLabel = new Label(p, SWT.NONE);
			errorLabel.setForeground(parent.getDisplay().getSystemColor(SWT.COLOR_RED));
			GridData elGD = new GridData(SWT.CENTER, SWT.NONE, true, false);
			elGD.horizontalSpan = 2;
			elGD.widthHint = 150;
			errorLabel.setLayoutData(elGD);
			
			new Label(p, SWT.NONE).setText("Variable Name:");
			
			nameBox = new Text(p, SWT.BORDER | SWT.SINGLE);
			GridData kcGD = new GridData(SWT.NONE, SWT.NONE, false, false);
			kcGD.widthHint = 100;
			nameBox.setLayoutData(kcGD);
			nameBox.setText(name);
			nameBox.addModifyListener(new ModifyListener() {
				public void modifyText(ModifyEvent e) {
					String text = nameBox.getText();
					int i = variables.indexOf(text + VARIABLE_SEPARATOR);
					while ((i > 0) && (variables.charAt(i-1) != '\t')) {
						i = variables.indexOf(text + VARIABLE_SEPARATOR, i+1);
					}
					if (!initialName.equals(text) && (i >= 0)) {
						errorLabel.setText("Name already in use.");
						getButton(Dialog.OK).setEnabled(false);
					} else {
						errorLabel.setText("");
						getButton(Dialog.OK).setEnabled(true);
					}
				}
			});
			
			new Label(p, SWT.NONE).setText("Value:");
			
			valueBox = new Text(p, SWT.BORDER | SWT.SINGLE);
			GridData resultGD = new GridData(SWT.FILL, SWT.NONE, true, false);
			resultGD.widthHint = IDialogConstants.ENTRY_FIELD_WIDTH;
			valueBox.setLayoutData(resultGD);
			valueBox.setText(value);
			return p;
		}
		
		protected void okPressed() {
			name = nameBox.getText();
			value = valueBox.getText();
			super.okPressed();
		}
	}
	
	Map<String, String> characters = new HashMap<String, String>();
	
	List characterList;
	Table variableList;
	Button editButton;
	
	public VariablesPreferencePage() {
		super();
	}

	public VariablesPreferencePage(String title) {
		super(title);
	}

	public VariablesPreferencePage(String title, ImageDescriptor image) {
		super(title, image);
	}

	protected Control createContents(Composite p) {
		noDefaultAndApplyButton();
		Composite parent = new Composite(p, SWT.NONE);
		parent.setLayout(new GridLayout(2, false));
		characterList = new List(parent, SWT.SINGLE | SWT.BORDER);
		GridData cListGD = new GridData(SWT.FILL, SWT.FILL, false, true);
		cListGD.widthHint = 100;
		characterList.setLayoutData(cListGD);
		characterList.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				updateVariableList();
			}
		});
		
		Composite innerParent = new Composite(parent, SWT.NONE);
		innerParent.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		innerParent.setLayout(new GridLayout(3, false));
		
		variableList = new Table(innerParent, SWT.SINGLE | SWT.FULL_SELECTION | SWT.BORDER);
		GridData varListGD = new GridData(SWT.FILL, SWT.FILL, true, true);
		varListGD.horizontalSpan = 3;
		variableList.setLayoutData(varListGD);
		TableColumn firstColumn = new TableColumn(variableList, SWT.NONE);
		firstColumn.setText("Variable Name");
		firstColumn.setWidth(100);
		TableColumn secondColumn = new TableColumn(variableList, SWT.NONE);
		secondColumn.setText("Value");
		secondColumn.setWidth(150);
		variableList.setHeaderVisible(true);
		variableList.addSelectionListener(new SelectionAdapter(){
			public void widgetSelected(SelectionEvent e) {
				editButton.setEnabled(true);
			}
		});
		
		Button addButton = new Button(innerParent, SWT.PUSH);
		addButton.setText("Add");
		addButton.setLayoutData(new GridData(SWT.FILL, SWT.NONE, true, false));
		addButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				EditVariableDialog dialog = new EditVariableDialog(getShell(), "", "", characters.get(characterList.getSelection()[0]));
				dialog.open();
				if (dialog.getReturnCode() == Window.OK) {
					addVariable(characterList.getSelection()[0], dialog.name + VARIABLE_SEPARATOR + dialog.value);
					int i = 0;
					while ((i < variableList.getItemCount()) && (variableList.getItem(i).getText(0).compareTo(dialog.value) < 0)) {
						i++;
					}
					TableItem newItem = new TableItem(variableList, SWT.NONE, i);
					newItem.setText(new String[]{dialog.name, dialog.value});
				}
			}
		});
		
		editButton = new Button(innerParent, SWT.PUSH);
		editButton.setText("Edit");
		editButton.setLayoutData(new GridData(SWT.FILL, SWT.NONE, true, false));
		editButton.setEnabled(false);
		editButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				if (variableList.getSelectionCount() == 1) {
					int selected = variableList.getSelectionIndex();
					TableItem toRemove = variableList.getItem(selected);
					String character = characterList.getSelection()[0];
					EditVariableDialog dialog = new EditVariableDialog(getShell(), toRemove.getText(0), toRemove.getText(1), characters.get(character));
					dialog.open();
					if (dialog.getReturnCode() == Window.OK) {
						removeVariable(character, toRemove.getText(0) + VARIABLE_SEPARATOR + toRemove.getText(1));
						addVariable(character, dialog.name + VARIABLE_SEPARATOR + dialog.value);
						toRemove.setText(new String[]{dialog.name, dialog.value});
					}
				}
			}
		});
		
		Button removeButton = new Button(innerParent, SWT.PUSH);
		removeButton.setText("Remove");
		removeButton.setLayoutData(new GridData(SWT.FILL, SWT.NONE, true, false));
		removeButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				if (variableList.getSelectionCount() == 1) {
					int selected = variableList.getSelectionIndex();
					TableItem toRemove = variableList.getItem(selected);
					String highlight = toRemove.getText(0) + VARIABLE_SEPARATOR + toRemove.getText(1);
					removeVariable(characterList.getSelection()[0], highlight);
					variableList.remove(selected);
					if (variableList.getItemCount() > selected) {
						variableList.setSelection(selected);
					} else {
						editButton.setEnabled(false);
					}
				}
			}
		});
		
		loadValues(doGetPreferenceStore().getString(PREF_VARIABLES));
		return parent;
	}

	public void init(IWorkbench workbench) {
		
	}
	
	void addVariable(String character, String variable) {
		String variables = characters.get(character);
		if (variables.length() == 0) {
			variables = variable;
		} else {
			variables += "\t" + variable;
		}
		characters.put(character, variables);
		if (EclipsefePlugin.DEBUG) {
			System.out.println("Variable added, variables for " + character + " are now: " + variables);
		}
	}
	
	void removeVariable(String character, String variable) {
		String variables = characters.get(character);
		int i = variables.indexOf(variable);
		while ((i > 0) && (variables.charAt(i-1) != '\t')) {
			i = variables.indexOf(variable, i+1);
		}
		if (i == 0) {
			if (variables.length() == variable.length()) {
				variables = "";
			} else {
				variables = variables.substring(variable.length()+1);
			}
		} else if (i > 0) {
			variables = variables.substring(0, i-1) + variables.substring(i+variable.length());
		}
		characters.put(character, variables);
	}
	
	void updateVariableList() {
		if (characterList.getSelectionCount() == 1) {
			variableList.removeAll();
			String variables = characters.get(characterList.getSelection()[0]);
			if (variables.length() > 0) {
				String[] varArray = variables.split("\t");
				Arrays.sort(varArray);
				for (String m : varArray) {
					TableItem item = new TableItem(variableList, SWT.NONE);
					item.setText(0, m.substring(0, m.indexOf(VARIABLE_SEPARATOR)));
					m = m.substring(m.indexOf(VARIABLE_SEPARATOR) + 1);
					item.setText(1, m);
				}
				editButton.setEnabled(false);
			}
		}
	}
	
	private void loadValues(String pref) {
		characters.clear();
		characterList.removeAll();
		String[] chars = pref.split("\n");
		Arrays.sort(chars);
		for (String c : chars) {
			if (c.length() > 0) {
				String name = c.substring(0, c.indexOf(':'));
				characters.put(name, c.substring(c.indexOf(':') + 1));
				characterList.add(name);
			}
		}
		variableList.clearAll();
	}
	
	public boolean performOk() {
		IPreferenceStore prefs = doGetPreferenceStore();
		StringBuilder results = new StringBuilder();
		for (String c : characters.keySet()) {
			results.append(c).append(':').append(characters.get(c)).append('\n');
		}
		prefs.setValue(PREF_VARIABLES, results.toString());
		return true;
	}
	
	protected void performDefaults() {
		super.performDefaults();
	}
	
	public static void setDefaults(IPreferenceStore prefs) {
		prefs.setDefault(PREF_VARIABLES, DEFAULT_VARIABLES);
	}
	
	protected IPreferenceStore doGetPreferenceStore() {
		return EclipsefePlugin.getDefault().getPreferenceStore();
	}
	
	public static Map<String, String> getVariables(String characterID) {
		IPreferenceStore prefs = EclipsefePlugin.getDefault().getPreferenceStore();
		String[] characters = prefs.getString(PREF_VARIABLES).split("\n");
		Map<String, String> vars = new HashMap<String, String>();
		String globalVariables = null;
		String charVariables = null;
		for (String character : characters) {
			if (character.length() > 0) {
				String name = character.substring(0, character.indexOf(':'));
				if (name.equals(characterID)) {
					charVariables = character.substring(character.indexOf(':') + 1);
				} else if (name.equals(MacroPreferencePage.GLOBAL_CHARACTER)) {
					globalVariables = character.substring(character.indexOf(':') + 1);
				}
			}
		}
		if ((globalVariables != null) && (globalVariables.length() > 0)) {
			addVariables(globalVariables.split("\t"), vars);
		}
		if ((charVariables != null) && (charVariables.length() > 0)) {
			addVariables(charVariables.split("\t"), vars);
		} else if (charVariables == null) {
			prefs.setValue(PREF_VARIABLES, prefs.getString(PREF_VARIABLES) + characterID + ":\n");
		}
		return vars;
	}
		
	private static void addVariables(String[] varList, Map<String, String> vars) {
		for (String m : varList) {
			String string = m.substring(0, m.indexOf(VARIABLE_SEPARATOR));
			m = m.substring(m.indexOf(VARIABLE_SEPARATOR) + 1);
			String value = m;
			
			vars.put(string, value);
		}
	}
	
	public static void setVariable(String charID, String name, String value) {
		IPreferenceStore prefs = EclipsefePlugin.getDefault().getPreferenceStore();
		String varlist = prefs.getString(PREF_VARIABLES);
		int start = varlist.indexOf(charID + ':');
		while ((start > 0) && (varlist.charAt(start-1) != '\n')) {
			start = varlist.indexOf(name + VARIABLE_SEPARATOR, start+1);
		}
		if (start == -1) {
			return;
		}
		start += charID.length() + 1;
		int end = varlist.indexOf('\n', start);
		if (end == -1) {
			return;
		}
		String vars = varlist.substring(start, end);
		int i = vars.indexOf(name + VARIABLE_SEPARATOR);
		while ((i > 0) && (vars.charAt(i-1) != '\t')) {
			i = vars.indexOf(name + VARIABLE_SEPARATOR, i+1);
		}
		if (i == -1) {
			if (vars.length() == 0) {
				vars = name + VARIABLE_SEPARATOR + value;
			} else {
				vars += '\t' + name + VARIABLE_SEPARATOR + value;
			}
		} else {
			i = vars.indexOf(VARIABLE_SEPARATOR, i);
			int j = vars.indexOf('\t', i);
			if (j == -1) {
				vars = vars.substring(0, i+1) + value;
			} else {
				vars = vars.substring(0, i+1) + value + vars.substring(j);
			}
		}
		varlist = varlist.substring(0, start) + vars + varlist.substring(end);
		prefs.setValue(PREF_VARIABLES, varlist);
	}
	
	public static boolean deleteVariable(String charID, String name) {
		IPreferenceStore prefs = EclipsefePlugin.getDefault().getPreferenceStore();
		String varlist = prefs.getString(PREF_VARIABLES);
		int start = varlist.indexOf(charID + ':');
		while ((start > 0) && (varlist.charAt(start-1) != '\n')) {
			start = varlist.indexOf(name + VARIABLE_SEPARATOR, start+1);
		}
		if (start == -1) {
			return false;
		}
		start += charID.length() + 1;
		int end = varlist.indexOf('\n', start);
		if (end == -1) {
			return false;
		}
		String vars = varlist.substring(start, end);
		int i = vars.indexOf(name + VARIABLE_SEPARATOR);
		while ((i > 0) && (vars.charAt(i-1) != '\t')) {
			i = vars.indexOf(name + VARIABLE_SEPARATOR, i+1);
		}
		if (i == -1) {
			return false;
		} else if (i == 0) {
			int j = vars.indexOf('\t');
			if (j == -1) {
				vars = "";
			} else {
				vars = vars.substring(j+1);
			}
		} else {
			int j = vars.indexOf('\t', i);
			if (j == -1) {
				vars = vars.substring(0, i-1);
			} else {
				vars = vars.substring(0, i-1) + vars.substring(j);
			}
		}
		varlist = varlist.substring(0, start) + vars + varlist.substring(end);
		prefs.setValue(PREF_VARIABLES, varlist);
		return true;
	}
	
	public void dispose() {
		super.dispose();
	}

}
