/*
 * Created on Nov 5, 2005
 */
package com.simutronics.eclipsefe.scripting;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.Path;
import org.eclipse.swt.custom.StyleRange;

import com.simutronics.eclipsefe.EclipsefePlugin;
import com.simutronics.eclipsefe.FEView;
import com.simutronics.eclipsefe.extensions.ICommandProcessor;
import com.simutronics.eclipsefe.extensions.IInputProcessor;
import com.simutronics.eclipsefe.preferences.FEPreferencePage;

/**
 * @author Adam Vartanian
 */
public class ScriptingEngine implements ICommandProcessor, IInputProcessor {
	
	private static Map<FEView, ScriptThread> executing = new HashMap<FEView, ScriptThread>();

	public void setInitializationData(IConfigurationElement config, String propertyName, Object data) throws CoreException {
	}

	public String processCommand(String command, FEView window) {
		String prefix = EclipsefePlugin.getDefault().getPreferenceStore().getString(FEPreferencePage.PREF_SCRIPT_PREFIX);
		if (!command.startsWith(prefix)) {
			return command;
		}
		window.fakeCommand(command);
		command = command.substring(prefix.length());
		String[] segments = command.split(" +", 2);
		Path scriptdir = new Path(EclipsefePlugin.getDefault().getPreferenceStore().getString(FEPreferencePage.PREF_SCRIPTS_DIRECTORY));
		if (scriptdir.toFile().exists()) {
			File script = scriptdir.addTrailingSeparator().append(segments[0]).addFileExtension("script").toFile();
			if (script.exists() && script.isFile()) {
				try {
					if (segments.length > 1) {
						executing.put(window, new ScriptThread(window, script, segments[1]));
					} else {
						executing.put(window, new ScriptThread(window, script, ""));
					}
					executing.get(window).start();
				} catch (IOException e) {
					window.displayError("Could not find script " + segments[0] + ".");
				}
			} else {
				window.displayError("Could not find script " + segments[0] + ".");
			}
		} else {
			window.displayError("Could not find script directory.");
		}
		return null;
	}
	
	public String processRawCommand(String string, FEView window) {
		return string;
	}

	public String processRawString(String string, FEView window) {
		if ((executing.get(window) != null) && executing.get(window).isAlive()) {
			executing.get(window).processRawString(string);
		}
		return string;
	}

	public String processParsedString(String string, String stream, List<StyleRange> styles, FEView window) {
		if ((executing.get(window) != null) && executing.get(window).isAlive()) {
			executing.get(window).processParsedString(string);
		}
		return string;
	}

}
